#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numbers

import numpy

import histbook.axis

class Channel(object):
    """Abstract class for graphical channels in a Vega-Lite plot."""

    def __init__(self, axis):
        self.axis = axis

class OverlayChannel(Channel):
    """Represents an overlayed axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".overlay({0})".format(self.axis)

class StackChannel(Channel):
    """Represents a stacked axis in a Vega-Lite plot."""

    def __init__(self, axis, order):
        super(StackChannel, self).__init__(axis)
        self.order = order

    def __repr__(self):
        if self.order is None:
            return ".stack({0})".format(self.axis)
        else:
            return ".stack({0}, order={1})".format(self.axis, repr(self.order))

class BesideChannel(Channel):
    """Represents a side-by-side axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".beside({0})".format(self.axis)

class BelowChannel(Channel):
    """Represents an above-and-below axis in a Vega-Lite plot."""

    def __repr__(self):
        return ".below({0})".format(self.axis)

class TerminalChannel(Channel):
    """Abstract class for the last graphical channel in a Vega-Lite plot."""

    def __init__(self, axis, profile, error, width, height, title, config, xscale, yscale, colorscale, shapescale):
        self.axis = axis
        self.profile = profile
        self.error = error
        self.width = width
        self.height = height
        self.title = title
        self.config = config
        self.xscale = xscale
        self.yscale = yscale
        self.colorscale = colorscale
        self.shapescale = shapescale

    def __repr__(self):
        args = [repr(self.axis)]
        if self.profile is not None:
            args.append("profile={0}".format(self.profile))
        if self.error is not False:
            args.append("error={0}".format(self.error))
        if self.width is not None:
            args.append("width={0}".format(repr(self.width)))
        if self.height is not None:
            args.append("height={0}".format(repr(self.height)))
        if self.title is not None:
            args.append("title={0}".format(repr(self.title)))
        if self.config is not None:
            args.append("config={0}".format(repr(self.config)))
        if self.xscale is not None:
            args.append("xscale={0}".format(repr(self.xscale)))
        if self.yscale is not None:
            args.append("yscale={0}".format(repr(self.yscale)))
        if self.colorscale is not None:
            args.append("colorscale={0}".format(repr(self.colorscale)))
        if self.shapescale is not None:
            args.append("shapescale={0}".format(repr(self.shapescale)))
        return ".{0}({1})".format(self._method, "".join(args))

class BarChannel(TerminalChannel):
    """Represents a bar axis in a Vega-Lite plot."""
    _method = "bar"

class StepChannel(TerminalChannel):
    """Represents a step axis in a Vega-Lite plot."""
    _method = "step"

class AreaChannel(TerminalChannel):
    """Represents an area axis in a Vega-Lite plot."""
    _method = "area"

class LineChannel(TerminalChannel):
    """Represents a line axis in a Vega-Lite plot."""
    _method = "line"

class MarkerChannel(TerminalChannel):
    """Represents a marker axis in a Vega-Lite plot."""
    _method = "marker"

class PlottingChain(object):
    """Mix-in for :py:class:`Hist <histbook.hist.Hist>` (as the first in a plotting chain) and :py:class:`Channels <histbook.vega.Channel>` in a plotting chain."""

    def __init__(self, source, item):
        if isinstance(source, PlottingChain):
            self._source = source._source
            self._chain = source._chain + (item,)
        else:
            self._source = source
            self._chain = (item,)

    def __repr__(self):
        return "".join(repr(x) for x in (self._source,) + self._chain)

    def __str__(self, indent="\n     ", paren=True):
        return ("(" if paren else "") + indent.join(repr(x) for x in (self._source,) + self._chain) + (")" if paren else "")

    def _singleaxis(self, axis):
        if axis is None:
            if len(self._source._group + self._source._fixed) == 1:
                axis, = self._source._group + self._source._fixed
            else:
                raise TypeError("histogram has more than one axis; one must be specified for plotting")
        return axis

    def _asaxis(self, axis):
        if axis is None:
            return None
        elif isinstance(axis, histbook.axis.Axis):
            return axis
        else:
            return self._source.axis[axis]

    def overlay(self, axis):
        """
        Display bins in ``axis`` overlaid on each other in different colors.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, OverlayChannel) for x in self._chain):
            raise TypeError("cannot overlay an overlay")
        return PlottingChain(self, OverlayChannel(self._asaxis(axis)))

    def stack(self, axis, order=None):
        """
        Display bins in ``axis`` stacked on one another in an area plot.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        order : iterable of strings
            stacking order of bins

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("cannot stack a stack")
        return PlottingChain(self, StackChannel(self._asaxis(axis), order))

    def beside(self, axis):
        """
        Display bins in ``axis`` next to each other horizontally.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, BesideChannel) for x in self._chain):
            raise TypeError("cannot split plots beside each other that are already split with beside (can do beside and below)")
        return PlottingChain(self, BesideChannel(self._asaxis(axis)))

    def below(self, axis):
        """
        Display bins in ``axis`` next to each other vertically.

        Parameters
        ----------
        axis : :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay

        Returns
        -------
        :py:class:`PlottingChain <histbook.vega.PlottingChain>`
        """
        if any(isinstance(x, BelowChannel) for x in self._chain):
            raise TypeError("cannot split plots below each other that are already split with below (can do beside and below)")
        return PlottingChain(self, BelowChannel(self._asaxis(axis)))

    def bar(self, axis=None, profile=None, error=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as bars on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable <histbook.vega.Plotable>`
        """
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable(self, BarChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def step(self, axis=None, profile=None, error=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as steps on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable <histbook.vega.Plotable>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable(self, StepChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def area(self, axis=None, profile=None, error=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as areas on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable <histbook.vega.Plotable>`
        """
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable(self, AreaChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def line(self, axis=None, profile=None, error=False, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as lines on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable <histbook.vega.Plotable>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable(self, LineChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, width, height, title, config, xscale, yscale, colorscale, shapescale))

    def marker(self, axis=None, profile=None, error=True, width=None, height=None, title=None, config=None, xscale=None, yscale=None, colorscale=None, shapescale=None):
        """
        Display bins in ``axis`` (if not the only axis) as markers on the horizontal axis.

        Parameters
        ----------
        axis : ``None``, :py:class:`Axis <histbook.axis.Axis>`, algebraic expression (lambda or string), or index position (integer)
            the axis to overlay; if ``None`` *(default)*, use the only axis in this :py:class:`Hist <histbook.hist.Hist>`

        profile : ``None``, :py:class:`profile <histbook.axis.profile>`, algebraic expression (lambda or string) or index position (integer)
            if ``None`` *(default)*, display bin counts; otherwise, display profile means (and errors on the mean)

        error : bool
            if ``True``, overlay error bars

        width, height, title, config, xscale, yscale, colorscale, shapescale : ``None`` or JSON
            graphical directives to pass to Vega-Lite

        Returns
        -------
        :py:class:`Plotable <histbook.vega.Plotable>`
        """
        if any(isinstance(x, StackChannel) for x in self._chain):
            raise TypeError("only area and bar can be stacked")
        if error and any(isinstance(x, (BesideChannel, BelowChannel)) for x in self._chain):
            raise NotImplementedError("error bars are currently incompatible with splitting beside or below")
        return Plotable(self, MarkerChannel(self._asaxis(self._singleaxis(axis)), self._asaxis(profile), error, width, height, title, config, xscale, yscale, colorscale, shapescale))

class Plotable(object):
    """Mix-in for :py:class:`Channels <histbook.vega.Channel>` and :py:class:`Combinations <histbook.vega.Combination>` that can be plotted."""

    def __init__(self, source, item):
        if isinstance(source, PlottingChain):
            self._source = source._source
            self._chain = source._chain + (item,)
        else:
            self._source = source
            self._chain = (item,)

    def __repr__(self):
        return "".join(repr(x) for x in (self._source,) + self._chain)

    def __str__(self, indent="\n     ", paren=True):
        return ("(" if paren else "") + indent.join(repr(x) for x in (self._source,) + self._chain) + (")" if paren else "")

    @property
    def _last(self):
        return self._chain[-1]

    def _data(self, prefix, varname):
        error = self._last.error
        baseline = isinstance(self._last, (StepChannel, AreaChannel))
        if isinstance(self._last.axis, (histbook.axis.bin, histbook.axis.intbin, histbook.axis.split)):
            if isinstance(self._last, BarChannel):
                xtype = "ordinal"
            else:
                xtype = "quantitative"
        else:
            xtype = "nominal"

        profile = self._last.profile
        if profile is None:
            profiles = ()
        else:
            profiles = (profile,)

        projected = self._source.project(*(x.axis for x in self._chain))
        table = projected.table(*profiles, count=(profile is None), error=error, recarray=False)

        projectedorder = [x for x in projected.axis if not isinstance(x, histbook.axis.ProfileAxis)]
        lastj = projectedorder.index(self._last.axis)

        data = []
        domains = {}
        if error and baseline:
            shifts = {}

        def recurse(j, content, row, base):
            if j == len(projectedorder):
                if base:
                    row = row + ((0.0, 0.0) if error else (0.0,))
                else:
                    row = row + tuple(float(x) for x in content)

                datum = dict(prefix + tuple(zip([varname + str(i) for i in range(len(row))], row)))
                if datum[varname + str(j)] > 0 or (self._last.yscale != "log" and not (isinstance(self._last.yscale, dict) and self._last.yscale.get("type", None) == "log")):
                    data.append(datum)

            else:
                axis = projectedorder[j]
                if axis not in domains:
                    domains[axis] = set()
                domains[axis].update(axis.keys(content))

                if isinstance(axis, histbook.axis.intbin):
                    axis = axis.bin()

                for i, (n, x) in enumerate(axis.items(content)):
                    if isinstance(n, histbook.axis.Interval):
                        if j == lastj and xtype == "quantitative":
                            if numpy.isfinite(n.low) and numpy.isfinite(n.high):
                                low = n.low
                                if baseline and isinstance(axis, (histbook.axis.bin, histbook.axis.split)) and n.low == axis.low:
                                    recurse(j + 1, x, row + (n.low,), True)
                                    low += 1e-10*(axis.high - axis.low)

                                recurse(j + 1, x, row + (low,), base)

                                if error and baseline:
                                    shifts[low] = 0.5*(n.low + n.high)

                                if baseline and isinstance(axis, (histbook.axis.bin, histbook.axis.split)) and n.high == axis.high:
                                    recurse(j + 1, x, row + (n.high,), True)

                        else:
                            recurse(j + 1, x, row + (str(n),), base)

                    elif isinstance(n, (bool, numpy.bool_, numpy.bool)):
                        recurse(j + 1, x, row + ("pass" if n else "fail",), base)

                    else:
                        recurse(j + 1, x, row + (str(n),), base)

        recurse(0, table, (), False)
        if error and baseline:
            for x in data:
                x[varname + str(lastj) + "c"] = shifts.get(x[varname + str(lastj)], x[varname + str(lastj)])

        return projectedorder, data, domains

    def _vegalite(self, axis, domains, varname):
        error = self._last.error
        baseline = isinstance(self._last, (StepChannel, AreaChannel))
        if isinstance(self._last.axis, (histbook.axis.bin, histbook.axis.intbin, histbook.axis.split)):
            if isinstance(self._last, BarChannel):
                xtype = "ordinal"
            else:
                xtype = "quantitative"
        else:
            xtype = "nominal"

        if isinstance(self._last, BarChannel):
            mark = "bar"
        elif isinstance(self._last, StepChannel):
            if xtype == "nominal":
                mark = "bar"
            else:
                mark = {"type": "line", "interpolate": "step-before"}
        elif isinstance(self._last, AreaChannel):
            if xtype == "nominal":
                mark = "bar"
            else:
                mark = {"type": "area", "interpolate": "step-before"}
        elif isinstance(self._last, LineChannel):
            mark = {"type": "line"}
        elif isinstance(self._last, MarkerChannel):
            mark = {"type": "point"}
        else:
            raise AssertionError(self._last)

        xtitle = self._last.axis.expr
        if self._last.profile is None:
            ytitle = "entries per bin"
        else:
            ytitle = self._last.profile.expr

        transform = []
        def makeorder(i, var, values):
            if len(values) == 1:
                return "if(datum.{0} === {1}, {2}, {3})".format(var, repr(values[0]), i, i + 1)
            elif len(values) > 1:
                return "if(datum.{0} === {1}, {2}, {3})".format(var, repr(values[0]), i, makeorder(i + 1, var, values[1:]))
            else:
                raise AssertionError(values)

        encoding = {"x": {"field": varname + str(axis.index(self._last.axis)), "type": xtype, "scale": {"zero": False}, "axis": {"title": xtitle}},
                    "y": {"field": varname + str(len(axis)), "type": "quantitative", "axis": {"title": ytitle}}}
        for channel in self._chain[:-1]:
            if isinstance(channel, OverlayChannel):
                overlayorder = [str(x) for x in sorted(domains[channel.axis])]
                encoding["color"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "legend": {"title": channel.axis.expr}, "scale": {"domain": overlayorder}}

            elif isinstance(channel, StackChannel):
                if channel.order is None:
                    stackorder = [str(x) for x in sorted(domains[channel.axis])]
                else:
                    stackorder = channel.order
                encoding["color"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "legend": {"title": channel.axis.expr}, "scale": {"domain": list(reversed(stackorder))}}
                encoding["y"]["aggregate"] = "sum"
                encoding["order"] = {"field": "stackorder", "type": "nominal"}
                transform.append({"calculate": makeorder(0, varname + str(axis.index(channel.axis)), stackorder), "as": "stackorder"})

            elif isinstance(channel, BesideChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["column"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            elif isinstance(channel, BelowChannel):
                # FIXME: sorting doesn't work??? https://github.com/vega/vega-lite/issues/2176
                encoding["row"] = {"field": varname + str(axis.index(channel.axis)), "type": "nominal", "header": {"title": channel.axis.expr}}

            else:
                raise AssertionError(channel)

        if self._last.xscale is not None and "x" in encoding:
            if isinstance(self._last.xscale, dict):
                encoding["x"]["scale"] = self._last.xscale
            else:
                encoding["x"]["scale"] = {"type": self._last.xscale}
        if self._last.yscale is not None and "y" in encoding:
            if isinstance(self._last.yscale, dict):
                encoding["y"]["scale"] = self._last.yscale
            else:
                encoding["y"]["scale"] = {"type": self._last.yscale}
        if self._last.colorscale is not None and "color" in encoding:
            if isinstance(self._last.colorscale, dict):
                encoding["color"]["scale"] = self._last.colorscale
            else:
                encoding["color"]["scale"] = {"type": self._last.colorscale}
        if self._last.shapescale is not None and "shape" in encoding:
            if isinstance(self._last.shapescale, dict):
                encoding["shape"]["scale"] = self._last.shapescale
            else:
                encoding["shape"]["scale"] = {"type": self._last.shapescale}

        if not error:
            return [mark], [encoding], [transform]

        else:
            if error and baseline:
                errx = varname + str(axis.index(self._last.axis)) + "c"
            else:
                errx = varname + str(axis.index(self._last.axis))

            encoding2 = {"x": {"field": errx, "type": "quantitative"},
                         "y": {"field": "error-down", "type": "quantitative"},
                         "y2": {"field": "error-up", "type": "quantitative"}}

            transform2 = [{"calculate": "datum.{0} - datum.{1}".format(varname + str(len(axis)), varname + str(len(axis) + 1)), "as": "error-down"},
                          {"calculate": "datum.{0} + datum.{1}".format(varname + str(len(axis)), varname + str(len(axis) + 1)), "as": "error-up"}]

            return [mark, "rule"], [encoding, encoding2], [transform, transform2]
        
    def vegalite(self):
        """Return the Vega-Lite JSON for this plot."""

        axis, data, domains = self._data((), "a")
        marks, encodings, transforms = self._vegalite(axis, domains, "a")

        if len(marks) == 1:
            return {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
                    "width": self._last.width,
                    "height": self._last.height,
                    "title": self._last.title,
                    "config": self._last.config,
                    "data": {"values": data},
                    "mark": marks[0],
                    "encoding": encodings[0],
                    "transform": transforms[0]}

        else:
            return {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
                    "width": self._last.width,
                    "height": self._last.height,
                    "title": self._last.title,
                    "config": self._last.config,
                    "data": {"values": data},
                    "layer": [{"mark": m, "encoding": e, "transform": t} for m, e, t in zip(marks, encodings, transforms)]}

    def to(self, fcn):
        """Call ``fcn`` on the Vega-Lite JSON for this plot."""
        return fcn(self.vegalite())

class Combination(object):
    """Abstract class for :py:class:`Plotables <histbook.vega.Plotable>` that have been combined as an overlay or side-by-side plots."""

    def __init__(self, *plotables):
        self._plotables = []
        for arg in plotables:           # first level: for varargs
            try:
                iter(arg)
            except:
                arg = [arg]
            for plotable in arg:        # second level: for iterators
                if not isinstance(plotable, Plotable):
                    raise TypeError("only Plotables can be combined with {0}".format(self.__class__.__name__))
                self._plotables.append(plotable)

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, ", ".join(repr(x) for x in self._plotables))

    def __str__(self, indent="\n    ", paren=False):
        return "{0}({1})".format(self.__class__.__name__, "".join(indent + x.__str__(indent + "    ", False) for x in self._plotables))

    def _varname(self, i):
        out = []
        i += 1
        while i > 0:
            out.append("abcdefghijklmnopqrstuvwxyz"[i % 26])
            i //= 26
        return "".join(reversed(out))

    def to(self, fcn):
        return fcn(self.vegalite())

class overlay(Combination):
    """:py:class:`Plotable <histbook.vega.Plotable>` overlaying two or more independently produced :py:class:`Plotables <histbook.vega.Plotable>`."""

    def vegalite(self):
        allaxis = []
        alldata = []
        alldomains = []
        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            axis, data, domains = plotable._data((("id", varname),), varname)
            allaxis.append(axis)
            alldata.extend(data)
            alldomains.append(domains)

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "layer": []}

        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], varname)
            thislayer = [{"filter": {"field": "id", "equal": varname}}]
            
            if len(marks) == 1:
                out["layer"].append({"mark": marks[0],
                                     "encoding": encodings[0],
                                     "transform": transforms[0] + thislayer})

            else:
                for m, e, t in zip(marks, encodings, transforms):
                    out["layer"].append({"mark": m, "encoding": e, "transform": t + thislayer})

        return out

class beside(Combination):
    """:py:class:`Plotable <histbook.vega.Plotable>` displaying two or more independently produced :py:class:`Plotables <histbook.vega.Plotable>` beside each other horizontally."""

    def __init__(self, *plotables):
        super(beside, self).__init__(*plotables)
        if any(isinstance(x, BesideChannel) for x in self._plotables):
            raise TypeError("cannot place plots beside each other that are already split with beside (can do beside and below)")

    def vegalite(self):
        allaxis = []
        alldata = []
        alldomains = []
        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            axis, data, domains = plotable._data((("id", varname),), varname)
            allaxis.append(axis)
            alldata.extend(data)
            alldomains.append(domains)

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "hconcat": []}

        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], varname)
            thislayer = [{"filter": {"field": "id", "equal": varname}}]

            if len(marks) == 1:
                out["hconcat"].append({"mark": marks[0], "encoding": encodings[0], "transform": transforms[0] + thislayer})

            else:
                out["hconcat"].append({"layer": [{"mark": m, "encoding": e, "transform": t} for m, e, t in zip(marks, encodings, transforms)]})

        return out

class below(Combination):
    """:py:class:`Plotable <histbook.vega.Plotable>` displaying two or more independently produced :py:class:`Plotables <histbook.vega.Plotable>` below each other vertically."""

    def __init__(self, *plotables):
        super(below, self).__init__(*plotables)
        if any(isinstance(x, BelowChannel) for x in self._plotables):
            raise TypeError("cannot place plots below each other that are already split with below (can do beside and below)")

    def vegalite(self):
        allaxis = []
        alldata = []
        alldomains = []
        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            axis, data, domains = plotable._data((("id", varname),), varname)
            allaxis.append(axis)
            alldata.extend(data)
            alldomains.append(domains)

        out = {"$schema": "https://vega.github.io/schema/vega-lite/v2.json",
               "data": {"values": alldata},
               "vconcat": []}

        for i, plotable in enumerate(self._plotables):
            varname = self._varname(i)
            marks, encodings, transforms = plotable._vegalite(allaxis[i], alldomains[i], varname)
            thislayer = [{"filter": {"field": "id", "equal": varname}}]

            if len(marks) == 1:
                out["vconcat"].append({"mark": marks[0], "encoding": encodings[0], "transform": transforms[0] + thislayer})

            else:
                out["vconcat"].append({"layer": [{"mark": m, "encoding": e, "transform": t} for m, e, t in zip(marks, encodings, transforms)]})

        return out



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import numpy

from histbook.axis import *
from histbook.hist import *

class TestProj(unittest.TestCase):
    def runTest(self):
        pass

    def test_project_bin(self):
        hxy = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False),
                   bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        hx = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        hy = Hist(bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        for i in range(1):
            hxy.fill(x=[0], y=[0])
            hx.fill(x=[0], y=[0])
            hy.fill(x=[0], y=[0])
        for i in range(2):
            hxy.fill(x=[0], y=[1])
            hx.fill(x=[0], y=[1])
            hy.fill(x=[0], y=[1])
        for i in range(4):
            hxy.fill(x=[1], y=[0])
            hx.fill(x=[1], y=[0])
            hy.fill(x=[1], y=[0])
        for i in range(8):
            hxy.fill(x=[1], y=[1])
            hx.fill(x=[1], y=[1])
            hy.fill(x=[1], y=[1])
        self.assertEqual(hxy._content.tolist(), [[[1], [2]], [[4], [8]]])
        self.assertEqual(hx._content.tolist(), [[3], [12]])
        self.assertEqual(hxy.project("x")._content.tolist(), [[3], [12]])
        self.assertEqual(hy._content.tolist(), [[5], [10]])
        self.assertEqual(hxy.project("y")._content.tolist(), [[5], [10]])

    def test_project_bin_weight(self):
        hxy = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False),
                   bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False), weight="w")
        hx = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False), weight="w")
        hy = Hist(bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False), weight="w")
        hxy.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8])
        hx.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8])
        hy.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8])
        self.assertEqual(hxy._content.tolist(), [[[1, 1], [2, 4]], [[4, 16], [8, 64]]])
        self.assertEqual(hx._content.tolist(), [[3, 5], [12, 80]])
        self.assertEqual(hxy.project("x")._content.tolist(), [[3, 5], [12, 80]])
        self.assertEqual(hy._content.tolist(), [[5, 17], [10, 68]])
        self.assertEqual(hxy.project("y")._content.tolist(), [[5, 17], [10, 68]])

    def test_project_bin_profile(self):
        hxy = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False),
                   bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False), profile("p"), weight="w")
        hx = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False), profile("p"), weight="w")
        hy = Hist(bin("y", 2, 0, 2, underflow=False, overflow=False, nanflow=False), profile("p"), weight="w")
        hxy.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8], p=[10, 20, 40, 80])
        hx.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8], p=[10, 20, 40, 80])
        hy.fill(x=[0, 0, 1, 1], y=[0, 1, 0, 1], w=[1, 2, 4, 8], p=[10, 20, 40, 80])
        self.assertEqual(hxy._content.tolist(), [[[10, 100, 1, 1], [40, 800, 2, 4]], [[160, 6400, 4, 16], [640, 51200, 8, 64]]])
        self.assertEqual(hx._content.tolist(), [[50, 900, 3, 5], [800, 57600, 12, 80]])
        self.assertEqual(hxy.project("x")._content.tolist(), [[50, 900, 3, 5], [800, 57600, 12, 80]])
        self.assertEqual(hy._content.tolist(), [[170, 6500, 5, 17], [680, 52000, 10, 68]])
        self.assertEqual(hxy.project("y")._content.tolist(), [[170, 6500, 5, 17], [680, 52000, 10, 68]])

    def test_project_groupby(self):
        hcx = Hist(groupby("c"), bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        hcx.fill(c=["one", "one", "one", "one", "two", "two"], x=[0, 1, 0, 1, 0, 1])
        self.assertEqual(set(hcx._content.keys()), set(["one", "two"]))
        self.assertEqual(set(hcx.project("c")._content.keys()), set(["one", "two"]))
        self.assertEqual(hcx.project("c").axis, (groupby("c"),))
        self.assertEqual(hcx.project("c")._content["one"].tolist(), [4])
        self.assertEqual(hcx.project("c")._content["two"].tolist(), [2])
        self.assertEqual(hcx.project("x")._content.tolist(), [[3], [3]])
        self.assertEqual(hcx.project("x").axis, (bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False),))

    def test_project_groupby_groupby(self):
        def tolist(obj):
            if isinstance(obj, dict):
                return dict((n, tolist(x)) for n, x in obj.items())
            else:
                return obj.tolist()

        hcdx = Hist(groupby("c"), groupby("d"), bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        hcdx.fill(c=["one", "one", "two", "two"], d=["uno", "dos", "un", "deux"], x=[0, 1, 0, 1])
        self.assertEqual(tolist(hcdx._content), {"two": {"dos": [[0], [0]], "un": [[1], [0]], "deux": [[0], [1]], "uno": [[0], [0]]}, "one": {"dos": [[0], [1]], "un": [[0], [0]], "deux": [[0], [0]], "uno": [[1], [0]]}})
        self.assertEqual(tolist(hcdx.project("c")._content), {"two": [2], "one": [2]})
        self.assertEqual(tolist(hcdx.project("c", "d").project("c")._content), {"two": [2], "one": [2]})
        self.assertEqual(tolist(hcdx.project("d")._content), {"dos": [1], "un": [1], "deux": [1], "uno": [1]})
        self.assertEqual(tolist(hcdx.project("c", "d").project("d")._content), {"dos": [1], "un": [1], "deux": [1], "uno": [1]})
        self.assertEqual(tolist(hcdx.project("c", "d")._content), {"two": {"dos": [0], "un": [1], "deux": [1], "uno": [0]}, "one": {"dos": [1], "un": [0], "deux": [0], "uno": [1]}})
        self.assertEqual(tolist(hcdx.project("d", "c")._content), {"two": {"dos": [0], "un": [1], "deux": [1], "uno": [0]}, "one": {"dos": [1], "un": [0], "deux": [0], "uno": [1]}})
        self.assertEqual(tolist(hcdx.project("x")._content), [[2], [2]])
        self.assertEqual(tolist(hcdx.project("c", "x").project("x")._content), [[2], [2]])
        self.assertEqual(tolist(hcdx.project("d", "x").project("x")._content), [[2], [2]])

    def test_project_away(self):
        empty = Hist()
        empty.fill(x=[1, 2, 3])
        self.assertEqual(empty._content.tolist(), [1])

        h = Hist(bin("x", 10, 0, 10))
        h.fill(x=[1, 2, 3])
        self.assertEqual(h.project()._content.tolist(), [3])

        empty = Hist(weight="y")
        empty.fill(x=[1, 2, 3], y=[0.1, 0.1, 0.1])
        self.assertEqual(empty._content.tolist(), [0.30000000000000004, 0.030000000000000006])

        h = Hist(bin("x", 10, 0, 10), weight="y")
        h.fill(x=[1, 2, 3], y=[0.1, 0.1, 0.1])
        self.assertEqual(h.project()._content.tolist(), [0.30000000000000004, 0.030000000000000006])

        hc = Hist(groupby("c"))
        hc.fill(c=["one", "one", "two"])
        self.assertEqual(hc.project()._content.tolist(), [3])

    def test_select_bin_axis(self):
        for underflow in False, True:
            for overflow in False, True:
                h = Hist(bin("x", 4, -1, 1, underflow=underflow, overflow=overflow))
                if underflow:
                    self.assertEqual(h.select("x < -1").axis["x"], bin("x", 0,   -1,   -1, underflow=underflow,    overflow=False, nanflow=False))
                self.assertEqual(h.select("x < -0.5").axis["x"],   bin("x", 1,   -1, -0.5, underflow=underflow,    overflow=False, nanflow=False))
                self.assertEqual(h.select("x < 0").axis["x"],      bin("x", 2,   -1,    0, underflow=underflow,    overflow=False, nanflow=False))
                self.assertEqual(h.select("x < 0.5").axis["x"],    bin("x", 3,   -1,  0.5, underflow=underflow,    overflow=False, nanflow=False))
                self.assertEqual(h.select("x < 1").axis["x"],      bin("x", 4,   -1,    1, underflow=underflow,    overflow=False, nanflow=False))

                self.assertEqual(h.select("x >= -1").axis["x"],    bin("x", 4,   -1,    1,     underflow=False, overflow=overflow, nanflow=False))
                self.assertEqual(h.select("x >= -0.5").axis["x"],  bin("x", 3, -0.5,    1,     underflow=False, overflow=overflow, nanflow=False))
                self.assertEqual(h.select("x >= 0").axis["x"],     bin("x", 2,    0,    1,     underflow=False, overflow=overflow, nanflow=False))
                self.assertEqual(h.select("x >= 0.5").axis["x"],   bin("x", 1,  0.5,    1,     underflow=False, overflow=overflow, nanflow=False))
                if overflow:
                    self.assertEqual(h.select("x >= 1").axis["x"], bin("x", 0,    1,    1,     underflow=False, overflow=overflow, nanflow=False))

                h = Hist(bin("x", 4, -1, 1, underflow=underflow, overflow=overflow, closedlow=False))
                if underflow:
                    self.assertEqual(h.select("x <= -1").axis["x"], bin("x", 0,   -1,   -1, underflow=underflow,    overflow=False, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x <= -0.5").axis["x"],   bin("x", 1,   -1, -0.5, underflow=underflow,    overflow=False, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x <= 0").axis["x"],      bin("x", 2,   -1,    0, underflow=underflow,    overflow=False, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x <= 0.5").axis["x"],    bin("x", 3,   -1,  0.5, underflow=underflow,    overflow=False, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x <= 1").axis["x"],      bin("x", 4,   -1,    1, underflow=underflow,    overflow=False, nanflow=False, closedlow=False))

                self.assertEqual(h.select("x > -1").axis["x"],      bin("x", 4,   -1,    1,     underflow=False, overflow=overflow, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x > -0.5").axis["x"],    bin("x", 3, -0.5,    1,     underflow=False, overflow=overflow, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x > 0").axis["x"],       bin("x", 2,    0,    1,     underflow=False, overflow=overflow, nanflow=False, closedlow=False))
                self.assertEqual(h.select("x > 0.5").axis["x"],     bin("x", 1,  0.5,    1,     underflow=False, overflow=overflow, nanflow=False, closedlow=False))
                if overflow:
                    self.assertEqual(h.select("x > 1").axis["x"],   bin("x", 0,    1,    1,     underflow=False, overflow=overflow, nanflow=False, closedlow=False))

    def test_select_bin_content(self):
        for nanflow in False, True:
            h = Hist(bin("x", 4, -1, 1, underflow=True, overflow=True, nanflow=nanflow))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x < -1")._content.tolist(), [[1]])
            self.assertEqual(h.select("x < -0.5")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x < 0")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x < 0.5")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x >= -1")._content.tolist(), [[2], [3], [4], [5], [6]])
            self.assertEqual(h.select("x >= -0.5")._content.tolist(), [[3], [4], [5], [6]])
            self.assertEqual(h.select("x >= 0")._content.tolist(), [[4], [5], [6]])
            self.assertEqual(h.select("x >= 0.5")._content.tolist(), [[5], [6]])
            self.assertEqual(h.select("x >= 1")._content.tolist(), [[6]])

            h = Hist(bin("x", 4, -1, 1, underflow=True, overflow=False, nanflow=nanflow))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x < -1")._content.tolist(), [[1]])
            self.assertEqual(h.select("x < -0.5")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x < 0")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x < 0.5")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x >= -1")._content.tolist(), [[2], [3], [4], [5]])
            self.assertEqual(h.select("x >= -0.5")._content.tolist(), [[3], [4], [5]])
            self.assertEqual(h.select("x >= 0")._content.tolist(), [[4], [5]])
            self.assertEqual(h.select("x >= 0.5")._content.tolist(), [[5]])

            h = Hist(bin("x", 4, -1, 1, underflow=False, overflow=True, nanflow=nanflow))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x < -0.5")._content.tolist(), [[1]])
            self.assertEqual(h.select("x < 0")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x < 0.5")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x >= -1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x >= -0.5")._content.tolist(), [[2], [3], [4], [5]])
            self.assertEqual(h.select("x >= 0")._content.tolist(), [[3], [4], [5]])
            self.assertEqual(h.select("x >= 0.5")._content.tolist(), [[4], [5]])
            self.assertEqual(h.select("x >= 1")._content.tolist(), [[5]])

            h = Hist(bin("x", 4, -1, 1, underflow=False, overflow=False, nanflow=nanflow))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x < -0.5")._content.tolist(), [[1]])
            self.assertEqual(h.select("x < 0")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x < 0.5")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x >= -1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x >= -0.5")._content.tolist(), [[2], [3], [4]])
            self.assertEqual(h.select("x >= 0")._content.tolist(), [[3], [4]])
            self.assertEqual(h.select("x >= 0.5")._content.tolist(), [[4]])

        for nanflow in False, True:
            h = Hist(bin("x", 4, -1, 1, underflow=True, overflow=True, nanflow=nanflow, closedlow=False))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x <= -1")._content.tolist(), [[1]])
            self.assertEqual(h.select("x <= -0.5")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x <= 0")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x <= 0.5")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x > -1")._content.tolist(), [[2], [3], [4], [5], [6]])
            self.assertEqual(h.select("x > -0.5")._content.tolist(), [[3], [4], [5], [6]])
            self.assertEqual(h.select("x > 0")._content.tolist(), [[4], [5], [6]])
            self.assertEqual(h.select("x > 0.5")._content.tolist(), [[5], [6]])
            self.assertEqual(h.select("x > 1")._content.tolist(), [[6]])

            h = Hist(bin("x", 4, -1, 1, underflow=True, overflow=False, nanflow=nanflow, closedlow=False))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x <= -1")._content.tolist(), [[1]])
            self.assertEqual(h.select("x <= -0.5")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x <= 0")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x <= 0.5")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x > -1")._content.tolist(), [[2], [3], [4], [5]])
            self.assertEqual(h.select("x > -0.5")._content.tolist(), [[3], [4], [5]])
            self.assertEqual(h.select("x > 0")._content.tolist(), [[4], [5]])
            self.assertEqual(h.select("x > 0.5")._content.tolist(), [[5]])

            h = Hist(bin("x", 4, -1, 1, underflow=False, overflow=True, nanflow=nanflow, closedlow=False))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x <= -0.5")._content.tolist(), [[1]])
            self.assertEqual(h.select("x <= 0")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x <= 0.5")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x > -1")._content.tolist(), [[1], [2], [3], [4], [5]])
            self.assertEqual(h.select("x > -0.5")._content.tolist(), [[2], [3], [4], [5]])
            self.assertEqual(h.select("x > 0")._content.tolist(), [[3], [4], [5]])
            self.assertEqual(h.select("x > 0.5")._content.tolist(), [[4], [5]])
            self.assertEqual(h.select("x > 1")._content.tolist(), [[5]])

            h = Hist(bin("x", 4, -1, 1, underflow=False, overflow=False, nanflow=nanflow, closedlow=False))
            h._content = numpy.arange(1, h.axis["x"].totbins + 1, dtype=numpy.float64).reshape(h.axis["x"].totbins, 1)
            self.assertEqual(h.select("x <= -0.5")._content.tolist(), [[1]])
            self.assertEqual(h.select("x <= 0")._content.tolist(), [[1], [2]])
            self.assertEqual(h.select("x <= 0.5")._content.tolist(), [[1], [2], [3]])
            self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x > -1")._content.tolist(), [[1], [2], [3], [4]])
            self.assertEqual(h.select("x > -0.5")._content.tolist(), [[2], [3], [4]])
            self.assertEqual(h.select("x > 0")._content.tolist(), [[3], [4]])
            self.assertEqual(h.select("x > 0.5")._content.tolist(), [[4]])

    def test_select_bin_bin(self):
        h = Hist(bin("x", 4, -1, 1, underflow=False, overflow=False, nanflow=False), bin("y", 4, -1, 1, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[0], y=[0])
        h.fill(x=[0], y=[0])
        h.fill(x=[0], y=[0])
        h.fill(x=[0], y=[0])
        h.fill(x=[0], y=[0.5])
        h.fill(x=[0], y=[0.5])
        h.fill(x=[0], y=[0.5])
        h.fill(x=[0.5], y=[0])
        h.fill(x=[0.5], y=[0])
        h.fill(x=[0.5], y=[0.5])
        self.assertEqual(h._content.tolist(), [[[0], [0], [0], [0]], [[0], [0], [0], [0]], [[0], [0], [4], [3]], [[0], [0], [2], [1]]])
        self.assertEqual(h.select("x >= 0")._content.tolist(), [[[0], [0], [4], [3]], [[0], [0], [2], [1]]])
        self.assertEqual(h.select("y >= 0")._content.tolist(), [[[0], [0]], [[0], [0]], [[4], [3]], [[2], [1]]])
        self.assertEqual(h.select("x >= 0 and y >= 0")._content.tolist(), [[[4], [3]], [[2], [1]]])
        self.assertEqual(h.select("x >= 0").select("y >= 0")._content.tolist(), [[[4], [3]], [[2], [1]]])

    def test_select_intbin(self):
        h = Hist(intbin("x", 5, 10))
        h.fill(x=range(15))
        self.assertEqual(h._content.tolist(), [[5], [1], [1], [1], [1], [1], [1], [4]])
        self.assertEqual(h.select("x < 6").axis["x"], intbin("x", 5, 5, overflow=False))
        self.assertEqual(h.select("x < 6")._content.tolist(), [[5], [1]])
        self.assertEqual(h.select("x <= 6").axis["x"], intbin("x", 5, 6, overflow=False))
        self.assertEqual(h.select("x <= 6")._content.tolist(), [[5], [1], [1]])
        self.assertEqual(h.select("x > 9").axis["x"], intbin("x", 10, 10, underflow=False))
        self.assertEqual(h.select("x > 9")._content.tolist(), [[1], [4]])
        self.assertEqual(h.select("x >= 9").axis["x"], intbin("x", 9, 10, underflow=False))
        self.assertEqual(h.select("x >= 9")._content.tolist(), [[1], [1], [4]])

        h = Hist(intbin("x", 5, 10, underflow=False, overflow=False))
        h.fill(x=range(15))
        self.assertEqual(h._content.tolist(), [[1], [1], [1], [1], [1], [1]])
        self.assertEqual(h.select("x < 6").axis["x"], intbin("x", 5, 5, underflow=False, overflow=False))
        self.assertEqual(h.select("x < 6")._content.tolist(), [[1]])
        self.assertEqual(h.select("x <= 6").axis["x"], intbin("x", 5, 6, underflow=False, overflow=False))
        self.assertEqual(h.select("x <= 6")._content.tolist(), [[1], [1]])
        self.assertEqual(h.select("x > 9").axis["x"], intbin("x", 10, 10, underflow=False, overflow=False))
        self.assertEqual(h.select("x > 9")._content.tolist(), [[1]])
        self.assertEqual(h.select("x >= 9").axis["x"], intbin("x", 9, 10, underflow=False, overflow=False))
        self.assertEqual(h.select("x >= 9")._content.tolist(), [[1], [1]])

    def test_select_split(self):
        h = Hist(split("x", (0, 1), underflow=True, overflow=True, closedlow=True))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[1], [2], [3], [4]])
        self.assertEqual(h.select("x < 0").axis["x"], split("x", 0, underflow=True, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 1").axis["x"], split("x", (0, 1), underflow=True, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 0")._content.tolist(), [[1]])
        self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("x >= 1").axis["x"], split("x", 1, underflow=False, overflow=True, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 0").axis["x"], split("x", (0, 1), underflow=False, overflow=True, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 1")._content.tolist(), [[3]])
        self.assertEqual(h.select("x >= 0")._content.tolist(), [[2], [3]])

        h = Hist(split("x", (0, 1), underflow=False, overflow=True, closedlow=True))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[2], [3], [4]])
        self.assertEqual(h.select("x < 1").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 1")._content.tolist(), [[2]])
        self.assertEqual(h.select("x >= 1").axis["x"], split("x", 1, underflow=False, overflow=True, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 0").axis["x"], split("x", (0, 1), underflow=False, overflow=True, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 1")._content.tolist(), [[3]])
        self.assertEqual(h.select("x >= 0")._content.tolist(), [[2], [3]])

        h = Hist(split("x", (0, 1), underflow=True, overflow=False, closedlow=True))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[1], [2], [4]])
        self.assertEqual(h.select("x < 0").axis["x"], split("x", 0, underflow=True, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 1").axis["x"], split("x", (0, 1), underflow=True, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 0")._content.tolist(), [[1]])
        self.assertEqual(h.select("x < 1")._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("x >= 0").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 0")._content.tolist(), [[2]])

        h = Hist(split("x", (0, 1), underflow=False, overflow=False, closedlow=True))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[2], [4]])
        self.assertEqual(h.select("x < 1").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x < 1")._content.tolist(), [[2]])
        self.assertEqual(h.select("x >= 0").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=True))
        self.assertEqual(h.select("x >= 0")._content.tolist(), [[2]])

        h = Hist(split("x", (0, 1), underflow=True, overflow=True, closedlow=False))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[1], [2], [3], [4]])
        self.assertEqual(h.select("x <= 0").axis["x"], split("x", 0, underflow=True, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 1").axis["x"], split("x", (0, 1), underflow=True, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 0")._content.tolist(), [[1]])
        self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("x > 1").axis["x"], split("x", 1, underflow=False, overflow=True, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 0").axis["x"], split("x", (0, 1), underflow=False, overflow=True, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 1")._content.tolist(), [[3]])
        self.assertEqual(h.select("x > 0")._content.tolist(), [[2], [3]])

        h = Hist(split("x", (0, 1), underflow=False, overflow=True, closedlow=False))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[2], [3], [4]])
        self.assertEqual(h.select("x <= 1").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 1")._content.tolist(), [[2]])
        self.assertEqual(h.select("x > 1").axis["x"], split("x", 1, underflow=False, overflow=True, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 0").axis["x"], split("x", (0, 1), underflow=False, overflow=True, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 1")._content.tolist(), [[3]])
        self.assertEqual(h.select("x > 0")._content.tolist(), [[2], [3]])

        h = Hist(split("x", (0, 1), underflow=True, overflow=False, closedlow=False))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[1], [2], [4]])
        self.assertEqual(h.select("x <= 0").axis["x"], split("x", 0, underflow=True, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 1").axis["x"], split("x", (0, 1), underflow=True, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 0")._content.tolist(), [[1]])
        self.assertEqual(h.select("x <= 1")._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("x > 0").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 0")._content.tolist(), [[2]])

        h = Hist(split("x", (0, 1), underflow=False, overflow=False, closedlow=False))
        h.fill(x=[-1, 0.5, 0.5, 2, 2, 2, numpy.nan, numpy.nan, numpy.nan, numpy.nan])
        self.assertEqual(h._content.tolist(), [[2], [4]])
        self.assertEqual(h.select("x <= 1").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x <= 1")._content.tolist(), [[2]])
        self.assertEqual(h.select("x > 0").axis["x"], split("x", (0, 1), underflow=False, overflow=False, nanflow=False, closedlow=False))
        self.assertEqual(h.select("x > 0")._content.tolist(), [[2]])

    def test_select_cut(self):
        h = Hist(cut("p"))
        h.fill(p=[True, True, False])
        self.assertEqual(h._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("p")._content.tolist(), [2])
        self.assertEqual(h.select("not p")._content.tolist(), [1])
        self.assertEqual(h.select("p")._fixed, ())
        self.assertEqual(h.select("p")._shape, (1,))
        self.assertEqual(h.select("not p")._fixed, ())
        self.assertEqual(h.select("not p")._shape, (1,))

        h = Hist(cut("x > 5"))
        h.fill(x=[8, 7, 3])
        self.assertEqual(h._content.tolist(), [[1], [2]])
        self.assertEqual(h.select("x > 5")._content.tolist(), [2])

    def test_select_groupby(self):
        h = Hist(groupby("c"))
        h.fill(c=["one", "two", "three", "two", "three", "three"])
        self.assertEqual(set(h.select("c == 'two'")._content.keys()), set(["two"]))
        self.assertEqual(set(h.select("c in {'two', 'three'}")._content.keys()), set(["two", "three"]))
        self.assertEqual(set(h.select("c > 'one'")._content.keys()), set(["two", "three"]))  # (alphabetically)
        self.assertEqual(set(h.select("c != 'two'")._content.keys()), set(["one", "three"]))
        self.assertEqual(set(h.select("not c == 'two'")._content.keys()), set(["one", "three"]))
        self.assertEqual(set(h.select("c != 'two' and c != 'three'")._content.keys()), set(["one"]))
        self.assertEqual(set(h.select("c == 'two' or c == 'three'")._content.keys()), set(["two", "three"]))

        h = Hist(groupby("c"), bin("x", 4, -1, 1, underflow=False, overflow=False, nanflow=False))
        h.fill(c=["one", "two", "three", "two", "three", "three"], x=[0, 0, 0, 0, 0, 0])
        self.assertEqual(h.select("c == 'two'")._content["two"].tolist(), [[0], [0], [2], [0]])
        self.assertEqual(h.select("c == 'two' and x >= 0")._content["two"].tolist(), [[2], [0]])
        self.assertEqual(h.select("x >= 0 and c == 'two'")._content["two"].tolist(), [[2], [0]])

    def test_select_groupbin(self):
        h = Hist(groupbin("x", 10))
        h.fill(x=[10, 15, 18, 20, 25])
        self.assertEqual(set(h._content.keys()), set([10.0, 20.0]))
        self.assertEqual(h._content[10.0].tolist(), [3])
        self.assertEqual(h._content[20.0].tolist(), [2])
        self.assertEqual(set(h.select("x < 20")._content.keys()), set([10.0]))
        self.assertEqual(h.select("x < 20")._content[10.0].tolist(), [3])
        self.assertEqual(h.select("not x >= 20")._content[10.0].tolist(), [3])

        h = Hist(groupbin("x", 10, closedlow=False))
        h.fill(x=[10, 15, 18, 20, 25])
        self.assertEqual(set(h._content.keys()), set([0.0, 10.0, 20.0]))
        self.assertEqual(h._content[0.0].tolist(), [1])
        self.assertEqual(h._content[10.0].tolist(), [3])
        self.assertEqual(h._content[20.0].tolist(), [1])
        self.assertEqual(set(h.select("x <= 20")._content.keys()), set([0.0, 10.0]))
        self.assertEqual(h.select("x <= 20")._content[0.0].tolist(), [1])
        self.assertEqual(h.select("x <= 20")._content[10.0].tolist(), [3])
        self.assertEqual(h.select("not x > 20")._content[10.0].tolist(), [3])

        h = Hist(groupbin("x", 10))
        h.fill(x=[15, 25, 25, 35, 35, 35])
        self.assertEqual(set(h._content.keys()), set([10.0, 20.0, 30.0]))
        self.assertEqual(set(h.select("x < 20 or x >= 30")._content.keys()), set([10.0, 30.0]))

    def test_rebin_split(self):
        h = Hist(split("x", (1, 2, 3)))
        h.fill(x=[0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)
        self.assertEqual(h._content.tolist(), [[1], [2], [4], [8], [16]])

        h2 = h.rebin("x", (2, 3))
        self.assertEqual(h2._content.tolist(), [[3], [4], [8], [16]])

        h3 = h.rebin("x", (1, 3))
        self.assertEqual(h3._content.tolist(), [[1], [6], [8], [16]])

        h4 = h.rebin("x", (1, 2))
        self.assertEqual(h4._content.tolist(), [[1], [2], [12], [16]])

        h5 = h.rebin("x", (1,))
        self.assertEqual(h5._content.tolist(), [[1], [14], [16]])

        h6 = h.rebin("x", (2,))
        self.assertEqual(h6._content.tolist(), [[3], [12], [16]])

        h7 = h.rebin("x", (3,))
        self.assertEqual(h7._content.tolist(), [[7], [8], [16]])

        h = Hist(split("x", (1, 2, 3), underflow=False, overflow=True))
        h.fill(x=[0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)
        self.assertEqual(h._content.tolist(), [[2], [4], [8], [16]])

        h2 = h.rebin("x", (2, 3))
        self.assertEqual(h2._content.tolist(), [[4], [8], [16]])

        h3 = h.rebin("x", (1, 3))
        self.assertEqual(h3._content.tolist(), [[6], [8], [16]])

        h4 = h.rebin("x", (1, 2))
        self.assertEqual(h4._content.tolist(), [[2], [12], [16]])

        h5 = h.rebin("x", (1,))
        self.assertEqual(h5._content.tolist(), [[14], [16]])

        h6 = h.rebin("x", (2,))
        self.assertEqual(h6._content.tolist(), [[12], [16]])

        h7 = h.rebin("x", (3,))
        self.assertEqual(h7._content.tolist(), [[8], [16]])

        h = Hist(split("x", (1, 2, 3), underflow=True, overflow=False))
        h.fill(x=[0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)
        self.assertEqual(h._content.tolist(), [[1], [2], [4], [16]])

        h2 = h.rebin("x", (2, 3))
        self.assertEqual(h2._content.tolist(), [[3], [4], [16]])

        h3 = h.rebin("x", (1, 3))
        self.assertEqual(h3._content.tolist(), [[1], [6], [16]])

        h4 = h.rebin("x", (1, 2))
        self.assertEqual(h4._content.tolist(), [[1], [2], [16]])

        h5 = h.rebin("x", (1,))
        self.assertEqual(h5._content.tolist(), [[1], [16]])

        h6 = h.rebin("x", (2,))
        self.assertEqual(h6._content.tolist(), [[3], [16]])

        h7 = h.rebin("x", (3,))
        self.assertEqual(h7._content.tolist(), [[7], [16]])

        h = Hist(split("x", (1, 2, 3), underflow=False, overflow=False))
        h.fill(x=[0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)
        self.assertEqual(h._content.tolist(), [[2], [4], [16]])

        h2 = h.rebin("x", (2, 3))
        self.assertEqual(h2._content.tolist(), [[4], [16]])

        h3 = h.rebin("x", (1, 3))
        self.assertEqual(h3._content.tolist(), [[6], [16]])

        h4 = h.rebin("x", (1, 2))
        self.assertEqual(h4._content.tolist(), [[2], [16]])

        h5 = h.rebin("x", (1,))
        self.assertEqual(h5._content.tolist(), [[16]])

        h6 = h.rebin("x", (2,))
        self.assertEqual(h6._content.tolist(), [[16]])

        h7 = h.rebin("x", (3,))
        self.assertEqual(h7._content.tolist(), [[16]])

    def test_rebin_split_bin(self):
        h = Hist(bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False), split("y", (1, 2, 3)))
        h.fill(x=[0]*(1+2+4+8+16) + [1]*(1+2+4+8+16), y=([0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)*2)
        self.assertEqual(h._content.tolist(), [[[1], [2], [4], [8], [16]], [[1], [2], [4], [8], [16]]])

        h2 = h.rebin("y", (2, 3))
        self.assertEqual(h2._content.tolist(), [[[3], [4], [8], [16]], [[3], [4], [8], [16]]])

        h3 = h.rebin("y", (1, 3))
        self.assertEqual(h3._content.tolist(), [[[1], [6], [8], [16]], [[1], [6], [8], [16]]])

        h4 = h.rebin("y", (1, 2))
        self.assertEqual(h4._content.tolist(), [[[1], [2], [12], [16]], [[1], [2], [12], [16]]])

        h5 = h.rebin("y", (1,))
        self.assertEqual(h5._content.tolist(), [[[1], [14], [16]], [[1], [14], [16]]])

        h6 = h.rebin("y", (2,))
        self.assertEqual(h6._content.tolist(), [[[3], [12], [16]], [[3], [12], [16]]])

        h7 = h.rebin("y", (3,))
        self.assertEqual(h7._content.tolist(), [[[7], [8], [16]], [[7], [8], [16]]])

        h = Hist(split("y", (1, 2, 3)), bin("x", 2, 0, 2, underflow=False, overflow=False, nanflow=False))
        h.fill(x=[0]*(1+2+4+8+16) + [1]*(1+2+4+8+16), y=([0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)*2)
        self.assertEqual(h._content.tolist(), [[[1], [1]], [[2], [2]], [[4], [4]], [[8], [8]], [[16], [16]]])

        h2 = h.rebin("y", (2, 3))
        self.assertEqual(h2._content.tolist(), [[[3], [3]], [[4], [4]], [[8], [8]], [[16], [16]]])

        h3 = h.rebin("y", (1, 3))
        self.assertEqual(h3._content.tolist(), [[[1], [1]], [[6], [6]], [[8], [8]], [[16], [16]]])

        h4 = h.rebin("y", (1, 2))
        self.assertEqual(h4._content.tolist(), [[[1], [1]], [[2], [2]], [[12], [12]], [[16], [16]]])

        h5 = h.rebin("y", (1,))
        self.assertEqual(h5._content.tolist(), [[[1], [1]], [[14], [14]], [[16], [16]]])

        h6 = h.rebin("y", (2,))
        self.assertEqual(h6._content.tolist(), [[[3], [3]], [[12], [12]], [[16], [16]]])

        h7 = h.rebin("y", (3,))
        self.assertEqual(h7._content.tolist(), [[[7], [7]], [[8], [8]], [[16], [16]]])

    def test_rebin_split_groupby(self):
        def tolist(obj):
            if isinstance(obj, dict):
                return dict((n, tolist(x)) for n, x in obj.items())
            else:
                return obj.tolist()

        h = Hist(groupby("c"), split("x", (1, 2, 3)))
        h.fill(c=["one"]*(1+2+4+8+16) + ["two"]*(1+2+4+8+16), x=([0] + [1]*2 + [2]*4 + [3]*8 + [numpy.nan]*16)*2)
        self.assertEqual(tolist(h._content), {"one": [[1], [2], [4], [8], [16]], "two": [[1], [2], [4], [8], [16]]})

        h2 = h.rebin("x", (2, 3))
        self.assertEqual(tolist(h2._content), {"one": [[3], [4], [8], [16]], "two": [[3], [4], [8], [16]]})

        h3 = h.rebin("x", (1, 3))
        self.assertEqual(tolist(h3._content), {"one": [[1], [6], [8], [16]], "two": [[1], [6], [8], [16]]})

        h4 = h.rebin("x", (1, 2))
        self.assertEqual(tolist(h4._content), {"one": [[1], [2], [12], [16]], "two": [[1], [2], [12], [16]]})

        h5 = h.rebin("x", (1,))
        self.assertEqual(tolist(h5._content), {"one": [[1], [14], [16]], "two": [[1], [14], [16]]})

        h6 = h.rebin("x", (2,))
        self.assertEqual(tolist(h6._content), {"one": [[3], [12], [16]], "two": [[3], [12], [16]]})

        h7 = h.rebin("x", (3,))
        self.assertEqual(tolist(h7._content), {"one": [[7], [8], [16]], "two": [[7], [8], [16]]})



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools

import histbook.expr

class CallGraphNode(object):
    """Represents a node in a graph of dependencies among subexpressions in an :py:class:`Expr <histbook.expr.Expr>`."""

    def __init__(self, goal):
        """Creates a node for a given goal (:py:class:`Expr <histbook.expr.Expr>`) without expanding to form a complete graph."""
        self.goal = goal
        self.clear()

    def clear(self):
        """Removes dependencies (``requires``) and dependent (``requiredby``) sets for this node."""
        self.requires = set()
        self.requiredby = set()
        self.numrequiredby = 0
        
    def __repr__(self):
        return "<CallGraphNode for {0}>".format(repr(str(self.goal)))

    def __hash__(self):
        return hash((CallGraphNode, self.goal))

    def __eq__(self, other):
        return isinstance(other, CallGraphNode) and self.goal == other.goal

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return self.goal < other.goal
        else:
            return self.__class__.__name__ < other.__class__.__name__

    def grow(self, table):
        """Builds a whole dependency graph, using ``table`` (dict) to identify identical nodes."""

        if self not in table:
            table[self] = self

        if isinstance(self.goal, (histbook.expr.Const, histbook.expr.Name, histbook.expr.Predicate)):
            pass

        elif isinstance(self.goal, histbook.expr.Call):
            for arg in self.goal.args:
                if not isinstance(arg, histbook.expr.Const):
                    node = CallGraphNode(arg)
                    if node not in table:
                        table[node] = node
                    else:
                        node = table[node]
                    self.requires.add(node)
                    node.requiredby.add(self)
                    node.grow(table)

        else:
            raise AssertionError(repr(self.goal))

        self.numrequiredby += 1

    def sources(self, table):
        """Returns the sources (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) of a dependency graph."""

        if isinstance(self.goal, histbook.expr.Const):
            return set()

        elif isinstance(self.goal, (histbook.expr.Name, histbook.expr.Predicate)):
            return set([self])

        elif isinstance(self.goal, histbook.expr.Call):
            out = set()
            for arg in self.goal.args:
                if not isinstance(arg, histbook.expr.Const):
                    node = table[CallGraphNode(arg)]
                    out.update(node.sources(table))
            return out

        else:
            raise NotImplementedError

    def rename(self, names):
        """Rename all :py:class:`Names <histbook.expr.Name>` and :py:class:`Predicates <histbook.expr.Predicate>` in the graph using ``names`` (dict) mapping old names to new names."""
        return self.goal.rename(names)

def totree(expr):
    """Simplifies ``expr`` (:py:class:`Expr <histbook.expr.Expr>`) to contain only constants (:py:class:`Const <histbook.expr.Const>`), names (:py:class:`Name <histbook.expr.Name>` and :py:class:`Predicate <histbook.expr.Predicate>`), and function calls (:py:class:`Call <histbook.expr.Call>`)."""

    def linear(fcn, args):
        if len(args) == 1:
            return args[0]
        return histbook.expr.Call(fcn, args[0], linear(fcn, args[1:]))

    def duplicates(fcn, args):
        if len(args) == 1:
            return args[0]

        uniques = set(args)
        if len(uniques) == len(args):
            return reduce(fcn, args)
        else:
            newargs = []
            for x in sorted(uniques):
                newargs.append(linear(fcn, (x,) * args.count(x)))
            return reduce(fcn, tuple(newargs))

    def reduce(fcn, args):
        if len(args) == 1:
            return args[0]

        left = args[:len(args) // 2]
        right = args[len(args) // 2:]

        if len(left) == 0:
            return reduce(fcn, right)
        if len(right) == 0:
            return reduce(fcn, left)

        if len(left) == 1:
            left, = left
        else:
            left = reduce(fcn, left)

        if len(right) == 1:
            right, = right
        else:
            right = reduce(fcn, right)

        return histbook.expr.Call(fcn, left, right)

    if isinstance(expr, (histbook.expr.Const, histbook.expr.Name, histbook.expr.Predicate)):
        return expr

    elif isinstance(expr, histbook.expr.Call):
        return histbook.expr.Call(expr.fcn, *(totree(x) for x in expr.args))

    elif isinstance(expr, histbook.expr.Relation):
        if expr.cmp == "==":
            return histbook.expr.Call("numpy.equal", totree(expr.left), totree(expr.right))

        elif expr.cmp == "!=":
            return histbook.expr.Call("numpy.not_equal", totree(expr.left), totree(expr.right))

        elif expr.cmp == "<":
            return histbook.expr.Call("numpy.less", totree(expr.left), totree(expr.right))

        elif expr.cmp == "<=":
            return histbook.expr.Call("numpy.less_equal", totree(expr.left), totree(expr.right))

        elif expr.cmp == "in":
            return histbook.expr.Call("numpy.isin", totree(expr.left), totree(expr.right))

        elif expr.cmp == "not in":
            return histbook.expr.Call("numpy.logical_not", histbook.expr.Call("numpy.isin", totree(expr.left), totree(expr.right)))

        else:
            raise AssertionError(repr(expr.cmp))

    elif isinstance(expr, histbook.expr.PlusMinus):
        out = None
        if len(expr.pos) > 0:
            out = reduce("numpy.add", tuple(totree(x) for x in expr.pos))

        if expr.const != expr.identity or out is None:
            if out is None:
                out = histbook.expr.Const(expr.const)
            else:
                out = histbook.expr.Call("numpy.add", out, histbook.expr.Const(expr.const))

        if len(expr.neg) > 0:
            out = histbook.expr.Call("numpy.subtract", out, reduce("numpy.add", tuple(totree(x) for x in expr.neg)))

        return out

    elif isinstance(expr, histbook.expr.TimesDiv):
        out = None
        if len(expr.pos) > 0:
            out = duplicates("numpy.multiply", tuple(totree(x) for x in expr.pos))

        if expr.const != expr.identity or out is None:
            if out is None:
                out = histbook.expr.Const(expr.const)
            else:
                out = histbook.expr.Call("numpy.multiply", out, histbook.expr.Const(expr.const))

        if len(expr.neg) > 0:
            out = histbook.expr.Call("numpy.true_divide", out, duplicates("numpy.multiply", tuple(totree(x) for x in expr.neg)))

        return out

    elif isinstance(expr, histbook.expr.LogicalOr):
        return reduce("numpy.logical_or", tuple(totree(x) for x in expr.args))

    elif isinstance(expr, histbook.expr.LogicalAnd):
        return reduce("numpy.logical_and", tuple(totree(x) for x in expr.args))

    else:
        raise AssertionError(expr)

class CallGraphGoal(CallGraphNode):
    """Goal node of a call graph: something to histogram."""
    def __init__(self, goal):
        super(CallGraphGoal, self).__init__(totree(goal))
        self.original = goal

def sources(goals, table):
    """Returns the sources (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) in a set of ``goals`` (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) given a ``table`` (dict) filled with :py:meth:`CallGraphNode.grow <histbook.instr.CallGraphNode.grow>`."""
    return functools.reduce(set.union, (x.sources(table) for x in goals), set())

def walkdown(sources):
    """Generator for an ordered walk from ``sources`` (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) to goals (:py:class:`CallGraphGoal <histbook.instr.CallGraphGoal>`) that steps through nodes required by the most goals first."""
    seen = set()
    def recurse(node):
        if node not in seen:
            seen.add(node)
            yield node

            pairs = [(x.numrequiredby, x) for x in node.requiredby]
            pairs.sort(reverse=True)
            for num, x in pairs:
                if all(y in seen for y in x.requires):
                    for y in recurse(x):
                        yield y

    for source in sources:
        for x in recurse(source):
            yield x

# def walkdown(sources):
#     """Generator for an ordered walk from sources (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) to goals (:py:class:`CallGraphGoal <histbook.instr.CallGraphGoal>`) that steps through nodes required by the most goals first."""
#     seen = set()
#     whenready = []
#     def recurse(node):
#         if node not in seen:
#             seen.add(node)
#             whenready.append(node)

#         notready = []
#         for trial in whenready:
#             if all(x in seen for x in trial.requires):
#                 yield trial
#             else:
#                 notready.append(trial)
#         del whenready[:]
#         whenready.extend(notready)

#         pairs = [(x.numrequiredby, x) for x in node.requiredby]
#         pairs.sort(reverse=True)
#         for num, x in pairs:
#             for y in recurse(x):
#                 yield y

#     for source in sources:
#         for x in recurse(source):
#             yield x

class Instruction(object):
    """Abstract class for instructions (ordered steps in the calculation)."""

class Param(Instruction):
    """Represents a request for data into the calculation."""

    def __init__(self, name, extern):
        self.name = name
        self.extern = extern

    def __repr__(self):
        return "Param({0}, {1})".format(repr(self.name), repr(self.extern))

    def __str__(self):
        return "{0} := {1} (external parameter)".format(self.name, repr(self.extern))

class Assign(Instruction):
    """Represents an assignment, creating a new variable to store the result of a function call."""

    def __init__(self, name, expr):
        self.name = name
        self.expr = expr

    def __repr__(self):
        return "Assign({0}, {1})".format(repr(self.name), repr(str(self.expr)))

    def __str__(self):
        return "{0} := {1}".format(self.name, str(self.expr))

class Export(Instruction):
    """Represents a result to export out of the calculation."""

    def __init__(self, name, goal):
        self.name = name
        self.goal = goal

    def __repr__(self):
        return "Export({0}, {1})".format(repr(self.name), repr(str(self.goal)))

    def __str__(self):
        return "export {0} as {1}".format(self.name, repr(str(self.goal)))

class Delete(Instruction):
    """Represents a subexpression that may now be safely deleted."""

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Delete({0})".format(repr(self.name))

    def __str__(self):
        return "delete {0}".format(self.name)

def instructions(sources, goals):
    """Returns an ordered sequence of instructions (:py:class:`Instruction <histbook.instr.Instruction>`), given a set of ``sources`` (:py:class:`CallGraphNode <histbook.instr.CallGraphNode>`) and a set of ``goals`` (:py:class:`CallGraphGoal <histbook.instr.CallGraphGoal>`)."""

    live = {}
    names = {}
    namenum = [0]
    def newname(node):
        name = "x{0}".format(namenum[0])
        namenum[0] += 1
        live[name] = node
        return name

    nodes = list(walkdown(sources))
    for i, node in enumerate(nodes):
        if isinstance(node.goal, histbook.expr.Const):
            pass

        elif isinstance(node.goal, (histbook.expr.Name, histbook.expr.Predicate)):
            name = newname(node)
            yield Param(name, node.goal.value)
            names[node.goal] = name

        elif isinstance(node.goal, histbook.expr.Call):
            name = newname(node)
            yield Assign(name, node.goal.rename(names))
            names[node.goal] = name

        else:
            raise NotImplementedError

        if node in goals:
            yield Export(name, node.goal)

        dead = []
        for n, x in live.items():
            if not any(x in nodes[j].requires for j in range(i + 1, len(nodes))):
                dead.append(n)
        for n in dead:
            del live[n]
            yield Delete(n)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import functools
import numbers
import sys

import numpy
COUNTTYPE = numpy.float64

import histbook.axis
import histbook.calc
import histbook.export
import histbook.expr
import histbook.proj
import histbook.instr
import histbook.vega

class _ChainedDict(object):
    def __init__(self, one, two):
        self._one = one
        self._two = two

    def __getitem__(self, n):
        if n in self._two:         # self._two is a real dict
            return self._two[n]    # and it has precedence
        else:
            return self._one[n]    # self._one might only have __getitem__
        
class Fillable(object):
    """Mix-in for objects with a ``fill`` method, like `Hist <histbook.hist.Hist>` and `Book <histbook.hist.Book>`."""

    @property
    def fields(self):
        """Names of fields that must be provided in the ``fill`` method."""

        if self._fields is None:
            table = {}
            goals = set(self._goals)

            for x in goals:
                x.clear()
            for x in goals:
                x.grow(table)
            
            fields = histbook.instr.sources(goals, table)

            self._instructions = self._streamline(0, list(histbook.instr.instructions(fields, goals)))
            self._fields = sorted(x.goal.value for x in fields)

        return self._fields

    def _showgoals(self):
        self.fields  # for the side-effect of creating self._instructions

        numbers = {}
        order = []
        def recurse(node):
            for x in node.requires:
                recurse(x)
            if node not in numbers:
                number = numbers[node] = len(numbers)
                order.append(node)
        for goal in sorted(self._goals):
            recurse(goal)
        print("goals:")
        print("------")
        for node in order:
            print("#{0:<3d} requires {1:<10s} requiredby {2:<10s} ({3} total) for {4}".format(numbers[node], " ".join(map(repr, sorted(numbers[x] for x in node.requires))), " ".join(map(repr, sorted(numbers[x] for x in node.requiredby))), node.numrequiredby, repr(str(node.goal))))
        print("")
        print("instructions:")
        print("-------------")
        for instruction in self._instructions:
            print(instruction)
        print("")
        
    def _fill(self, arrays):
        self.fields  # for the side-effect of creating self._instructions

        length = None
        symbols = {}
        for instruction in self._instructions:
            if isinstance(instruction, histbook.instr.Param):
                try:
                    array = arrays[instruction.extern]
                except KeyError:
                    raise ValueError("required field {0} not found in fill arguments".format(repr(instruction.extern)))

                if not isinstance(array, numpy.ndarray):
                    array = numpy.array(array)
                if array.shape == ():
                    array.shape = (1,)

                if length is None:
                    length = len(array)
                elif length != len(array):
                    raise ValueError("array {0} has len {1} but other arrays have len {2}".format(repr(instruction.extern), len(array), length))

                symbols[instruction.name] = array

            elif isinstance(instruction, histbook.instr.Assign):
                symbols[instruction.name] = histbook.calc.calculate(instruction.expr, symbols)

            elif isinstance(instruction, histbook.instr.Export):
                data = symbols[instruction.name]
                for i, j in instruction.destination:
                    self._destination[i][j] = data

            elif isinstance(instruction, histbook.instr.Delete):
                del symbols[instruction.name]

            else:
                raise AssertionError(instruction)

        return length

class Book(collections.MutableMapping, Fillable):
    """
    A collection of histograms (:py:class:`Hist <histbook.hist.Hist>`) that can be filled with a single ``fill`` call.

    Behaves like a dict (item assignment, ``keys``, ``values``).
    """

    def __init__(self, hists={}, **more):
        u"""
        Parameters
        ----------
        hists : dict of str \u2192 :py:class:`Hist <histbook.hist.Hist>`
            initial histograms

        **more : dict of str \u2192 :py:class:`Hist <histbook.hist.Hist>`
            more initial histograms
        """
        self._fields = None
        self._hists = collections.OrderedDict()
        for n, x in hists.items():
            self._hists[n] = x
        for n, x in more.items():
            self._hists[n] = x

    def __repr__(self):
        return "Book({0} histogram{1})".format(len(self), "" if len(self) == 1 else "s")

    def __str__(self):
        return "Book({" + ",\n      ".join("{0}: {1}".format(repr(n), repr(x)) for n, x in self.items()) + "})"

    def __len__(self):
        return len(self._hists)

    def __contains__(self, name):
        return name in self._hists

    def __getitem__(self, name):
        return self._hists[name]

    def __setitem__(self, name, value):
        if isinstance(value, Book):
            for n, x in value.items():
                self._hists[name + "/" + n] = x.copyonfill()
                self._fields = None
        elif isinstance(value, Hist):
            self._hists[name] = value.copyonfill()
            self._fields = None
        else:
            raise TypeError("histogram books can only be filled with histograms or other histogram books, not {0}".format(type(value)))

    def __delitem__(self, name):
        del self._hists[name]

    def __iter__(self):
        if sys.version_info[0] < 3:
            return self._hists.iterkeys()
        else:
            return self._hists.keys()

    def keys(self):
        return self._hists.keys()

    def values(self):
        return self._hists.values()

    @property
    def _goals(self):
        return functools.reduce(set.union, (x._goals for x in self.values()))

    def _streamline(self, i, instructions):
        self._destination = []
        for i, x in enumerate(self._hists.values()):
            self._destination.append(x._destination[0])
            x._streamline(i, instructions)
        return instructions

    def fill(self, arrays=None, **more):
        u"""
        Fill the histogram: identify bins for independent variables, increase their counts by ``1`` or ``weight``, and increment any profile (dependent variable) means and errors in the means.

        All arrays must have the same length (one-dimensional shape). Numbers are treated as one-element arrays.

        All histograms in the book are filled with the same inputs.

        Parameters
        ----------
        arrays : dict \u2192 Numpy array or number
            field values to use in the calculation of independent and dependent variables (axes)

        **more : dict \u2192 Numpy array or number
            more field values
        """
        if arrays is None:
            arrays = more
        elif len(more) == 0:
            pass
        else:
            arrays = _ChainedDict(arrays, more)

        for x in self._hists.values():
            x._prefill()
        length = self._fill(arrays)
        for x in self._hists.values():
            x._postfill(arrays, length)

    def __add__(self, other):
        if not isinstance(other, Book):
            raise TypeError("histogram Books can only be added to other histogram Books")

        out = Book(self._hists)
        for n, x in other.items():
            if n in out:
                out[n] += x
            else:
                out[n] = x

        return out

    def __iadd__(self, other):
        if not isinstance(other, Book):
            raise TypeError("histogram Books can only be added to other histogram Books")

        for n, x in other.items():
            if n in self:
                self[n] += x
            else:
                self[n] = x

        return self

    @staticmethod
    def group(by="source", **books):
        """
        Combine histograms, maintaining their distinctiveness by adding a new categorical axis to each.

        To combine histograms by adding bins, just use the ``+`` operator.

        Parameters
        ----------
        by : string
            name of the new axis (must not already exist)

        **books : :py:class:`Book <histbook.hist.Book>`
            books to combine (histograms with the same names must have the same axes)
        """
        if any(not isinstance(x, Book) for x in books.values()):
            raise TypeError("only histogram Books can be grouped")
        out = Book()
        for n in functools.reduce(set.union, (set(x.keys()) for x in books.values())):
            out._hists[n] = Hist.group(by=by, **dict((name, book[n]) for name, book in books.items() if n in book.keys()))
        return out

class Hist(Fillable, histbook.proj.Projectable, histbook.export.Exportable, histbook.vega.PlottingChain):
    @property
    def _source(self):
        return self

    @property
    def _chain(self):
        return ()

    def weight(self, expr):
        """Returns a copy of this histogram with ``expr`` as weights (for fluent construction)."""
        return Hist(*[x.relabel(x._original) for x in self._group + self._fixed + self._profile], weight=expr, defs=self._defs)

    @staticmethod
    def _copycontent(content):
        if content is None:
            return None
        elif isinstance(content, numpy.ndarray):
            return content.copy()
        else:
            return dict((n, Hist._copycontent(x)) for n, x in content.items())

    def copy(self):
        """Return an immediate copy of the histogram."""
        out = self.__class__.__new__(self.__class__)
        out.__dict__.update(self.__dict__)
        out._content = Hist._copycontent(self._content)
        return out

    def copyonfill(self):
        """Return a copy of the histogram whose content is copied if filled."""
        out = self.__class__.__new__(self.__class__)
        out.__dict__.update(self.__dict__)
        out._copyonfill = True
        return out

    def __init__(self, *axis, **opts):
        u"""
        Parameters
        ----------
        *axis : :py:class:`Axis <histbook.axis.Axis>`
            axis or axes that define the independent and dependent variables of the histogram

        Keyword Arguments
        -----------------
        weight : ``None``, algebraic expression (lambda or string), or number
            if ``None`` *(default)*, data will be filled with weight ``1``; if an expression, weights are computed from the expression; if a number, weights are constant

        defs : ``None`` or dict of str \u2192 algebraic expression (lambda or string) or :py:class:`Expr <histbook.expr.Expr>`
            if not ``None``, definitions to use when computing expressions

        fill : ``None``, single Numpy array or dict of str \u2192 Numpy arrays
            if not ``None``, data to immediately fill after constructing the histogram; single Numpy array is only permitted if there's only one field
        """
        weight = opts.pop("weight", None)
        defs = opts.pop("defs", {})
        fill = opts.pop("fill", None)
        if len(opts) > 0:
            raise TypeError("unrecognized options for Hist: {0}".format(" ".join(opts)))

        self._defs = defs
        self._group = []
        self._fixed = []
        self._profile = []

        newaxis = []
        for old in axis:
            if isinstance(old, histbook.axis._nullaxis) or (hasattr(old, "_original") and hasattr(old, "_parsed")):
                newaxis.append(old)
            else:
                expr, label = histbook.expr.Expr.parse(old._expr, defs=defs, returnlabel=True)
                new = old.relabel(label)
                new._original = old._expr
                new._parsed = expr
                newaxis.append(new)

        self._goals = set()
        self._destination = [[]]
        self._lookup = {}
        def dest(goals):
            self._goals.update(set(goals))
            for goal in goals:
                if goal.goal not in self._lookup:
                    self._lookup[goal.goal] = []
                self._lookup[goal.goal].append(len(self._destination[0]))
                self._destination[0].append(None)

        dictindex = 0
        for new in newaxis:
            if isinstance(new, histbook.axis.GroupAxis):
                self._group.append(new)
                new._dictindex = dictindex
                dictindex += 1
                dest(new._goals(new._parsed))

        self._shape = []
        for new in newaxis:
            if isinstance(new, histbook.axis.FixedAxis):
                self._fixed.append(new)
                new._shapeindex = len(self._shape)
                self._shape.append(new.totbins)
                if not isinstance(new, histbook.axis._nullaxis):
                    dest(new._goals(new._parsed))

        self._shape.append(0)
        for new in newaxis:
            if isinstance(new, histbook.axis.ProfileAxis):
                self._profile.append(new)
                new._sumwxindex = self._shape[-1]
                new._sumwx2index = self._shape[-1] + 1
                self._shape[-1] += 2
                dest(new._goals(new._parsed))

        if weight is None:
            self._weightoriginal, self._weightparsed, self._weightlabel = None, None, None
            self._sumwindex = self._shape[-1]
            self._shape[-1] += 1
        
        else:
            self._weightoriginal = weight
            if isinstance(weight, (numbers.Real, numpy.integer, numpy.floating)):
                self._weightparsed, self._weightlabel = histbook.expr.Const(weight), str(weight)
            else:
                self._weightparsed, self._weightlabel = histbook.expr.Expr.parse(weight, defs=self._defs, returnlabel=True)
            self._sumwindex = self._shape[-1]
            self._sumw2index = self._shape[-1] + 1
            self._shape[-1] += 2
            dest([histbook.instr.CallGraphGoal(self._weightparsed),
                  histbook.instr.CallGraphGoal(histbook.expr.Call("numpy.multiply", self._weightparsed, self._weightparsed))])
            
        self._group = tuple(self._group)
        self._fixed = tuple(self._fixed)
        self._profile = tuple(self._profile)

        self._weight = weight
        self._shape = tuple(self._shape)
        self._content = None
        self._fields = None
        self._copyonfill = False

        if fill is not None:
            if not isinstance(fill, dict):
                if len(self._group + self._fixed + self._profile) == 1:
                    fill = {str((self._group + self._fixed + self._profile)[0]._parsed): fill}
                else:
                    raise TypeError("fill must be a dict for histograms of more than one axis")
            self.fill(fill)

    def __repr__(self, indent=", "):
        out = [repr(x) for x in self._group + self._fixed + self._profile]
        if self._weightlabel is not None:
            out.append("weight={0}".format(repr(self._weightlabel)))
        if len(self._defs) > 0:
            out.append("defs={" + ", ".join("{0}: {1}".format(repr(n), repr(str(x)) if isinstance(x, histbook.expr.Expr) else repr(x)) for n, x in self._defs.items()) + "}")
        return "Hist(" + indent.join(out) + ")"

    def __str__(self):
        return self.__repr__(",\n     ")

    @property
    def shape(self):
        """Shape of the Numpy array defining the content of the fixed-memory axes (:py:class:`FixedAxis <histbook.axis.FixedAxis>`) only."""
        return self._shape

    def _streamline(self, i, instructions):
        for instruction in instructions:
            if isinstance(instruction, histbook.instr.Export):
                if not hasattr(instruction, "destination"):
                    instruction.destination = []
                if instruction.goal in self._lookup:
                    for j in self._lookup[instruction.goal]:
                        instruction.destination.append((i, j))

        return instructions

    def fill(self, arrays=None, **more):
        u"""
        Fill the histogram: identify bins for independent variables, increase their counts by ``1`` or ``weight``, and increment any profile (dependent variable) means and errors in the means.

        All arrays must have the same length (one-dimensional shape). Numbers are treated as one-element arrays.

        Parameters
        ----------
        arrays : dict \u2192 Numpy array or number
            field values to use in the calculation of independent and dependent variables (axes)

        **more : dict \u2192 Numpy array or number
            more field values
        """
        if self._copyonfill:
            self._content = Hist._copycontent(self._content)
            self._copyonfill = False

        if arrays is None:
            arrays = more
        elif len(more) == 0:
            pass
        else:
            arrays = _ChainedDict(arrays, more)

        self._prefill()
        length = self._fill(arrays)
        self._postfill(arrays, length)

    def _prefill(self):
        if self._content is None:
            if len(self._group) == 0:
                self._content = numpy.zeros(self._shape, dtype=COUNTTYPE)
            else:
                self._content = {}

    def _postfill(self, arrays, length):
        j = len(self._group)
        step = 0
        indexes = None
        for axis in self._fixed:
            if step == 0:
                indexes = self._destination[0][j]
            elif step == 1:
                indexes = indexes.copy()
            if step > 0:
                numpy.multiply(indexes, self._shape[axis._shapeindex], indexes)
                numpy.add(indexes, self._destination[0][j], indexes)
            j += 1
            step += 1

        axissumx, axissumx2 = [], []
        for axis in self._profile:
            axissumx.append(self._destination[0][j])
            axissumx2.append(self._destination[0][j + 1])
            j += 2

        if self._weightparsed is None:
            weight = 1
            weight2 = None
        elif isinstance(self._weightparsed, histbook.expr.Const):
            weight = numpy.ones(length) * self._weightparsed.value
            weight2 = numpy.ones(length) * self._weightparsed.value**2
        else:
            weight = self._destination[0][j]
            weight2 = self._destination[0][j + 1]
            selection = numpy.isnan(weight)
            if selection.any():
                weight = weight.copy()
                weight2 = weight2.copy()
                weight[selection] = 0.0
                weight2[selection] = 0.0

        def fillblock(content, indexes, axissumx, axissumx2, weight, weight2):
            for sumx, sumx2, axis in zip(axissumx, axissumx2, self._profile):
                if indexes is None:
                    indexes = numpy.ma.zeros(len(sumx), dtype=histbook.calc.INDEXTYPE)
                if weight2 is not None:
                    selection = numpy.ma.getmask(indexes)
                    if selection is not numpy.ma.nomask:
                        selection = numpy.bitwise_not(selection)
                        weight = weight[selection]
                        weight2 = weight2[selection]
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, axis._sumwxindex], indexes.compressed(), sumx * weight)
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, axis._sumwx2index], indexes.compressed(), sumx2 * weight)

            if weight2 is None:
                if indexes is None:
                    content.reshape((-1, self._shape[-1]))[:, self._sumwindex] += (1 if length is None else length) * weight
                else:
                    numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumwindex], indexes.compressed(), weight)
            else:
                if indexes is None:
                    indexes = numpy.ma.zeros(len(weight), dtype=histbook.calc.INDEXTYPE)
                selection = numpy.ma.getmask(indexes)
                if selection is not numpy.ma.nomask:
                    selection = numpy.bitwise_not(selection)
                    weight = weight[selection]
                    weight2 = weight2[selection]
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumwindex], indexes.compressed(), weight)
                numpy.add.at(content.reshape((-1, self._shape[-1]))[:, self._sumw2index], indexes.compressed(), weight2)

        def filldict(j, content, indexes, axissumx, axissumx2, weight, weight2, allselection):
            if j == len(self._group):
                fillblock(content, indexes, axissumx, axissumx2, weight, weight2)

            else:
                uniques, inverse = self._destination[0][j]
                for idx, unique in enumerate(uniques):
                    if allselection is None:
                        selection = (inverse == idx)
                    else:
                        selection = (inverse[allselection] == idx)
                    
                    if unique not in content:
                        if j + 1 == len(self._group):
                            content[unique] = numpy.zeros(self._shape, dtype=COUNTTYPE)
                        else:
                            content[unique] = {}

                    subcontent = content[unique]
                    if indexes is None:
                        subindexes = numpy.ma.zeros(numpy.count_nonzero(selection), dtype=histbook.calc.INDEXTYPE)
                    else:
                        subindexes = indexes[selection]
                    subaxissumx = [x[selection] for x in axissumx]
                    subaxissumx2 = [x[selection] for x in axissumx2]
                    if weight2 is None:
                        subweight, subweight2 = weight, weight2
                    else:
                        subweight = weight[selection]
                        subweight2 = weight2[selection]

                    if allselection is None:
                        suballselection = selection
                    else:
                        suballselection = allselection.copy()
                        suballselection[inverse != idx] = False

                    filldict(j + 1, subcontent, subindexes, subaxissumx, subaxissumx2, subweight, subweight2, suballselection)

        filldict(0, self._content, indexes, axissumx, axissumx2, weight, weight2, None)
            
        for j in range(len(self._destination[0])):
            self._destination[0][j] = None

    def __add__(self, other):
        if not isinstance(other, Hist):
            raise TypeError("histograms can only be added to other histograms")

        if self._group + self._fixed + self._profile != other._group + other._fixed + other._profile:
            raise TypeError("histograms can only be added to other histograms with the same axis specifications")

        def add(selfcontent, othercontent):
            if selfcontent is None and othercontent is None:
                return None

            elif selfcontent is None:
                return Hist._copycontent(othercontent)

            elif othercontent is None:
                return Hist._copycontent(selfcontent)

            elif isinstance(selfcontent, numpy.ndarray) and isinstance(othercontent, numpy.ndarray):
                return selfcontent + othercontent

            else:
                assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
                out = {}
                for n in selfcontent:
                    if n in othercontent:
                        out[n] = add(selfcontent[n], othercontent[n])
                    else:
                        out[n] = selfcontent[n]
                for n in othercontent:
                    if n not in selfcontent:
                        out[n] = othercontent[n]
                return out

        out = self.__class__.__new__(self.__class__)
        out.__dict__.update(self.__dict__)
        out._content = add(self._content, other._content)
        return out

    def __iadd__(self, other):
        if not isinstance(other, Hist):
            raise TypeError("histograms can only be added to other histograms")

        if self._group + self._fixed + self._profile != other._group + other._fixed + other._profile:
            raise TypeError("histograms can only be added to other histograms with the same axis specifications")

        def add(selfcontent, othercontent):
            assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
            for n in selfcontent:
                if n in othercontent:
                    if isinstance(selfcontent[n], numpy.ndarray):
                        selfcontent[n] += othercontent[n]
                    else:
                        add(selfcontent[n], othercontent[n])
            for n in othercontent:
                if n not in selfcontent:
                    selfcontent[n] = Hist._copycontent(othercontent[n])

        if other._content is None:
            pass

        elif self._content is None:
            self._content = Hist._copycontent(other._content)

        elif isinstance(self._content, numpy.ndarray):
            self._content += other._content

        else:
            add(self._content, other._content)

        return self

    @staticmethod
    def group(by="source", **hists):
        """
        Combine histograms, maintaining their distinctiveness by adding a new categorical axis to each.

        To combine histograms by adding bins, just use the ``+`` operator.

        Parameters
        ----------
        by : string
            name of the new axis (must not already exist)

        **hists : :py:class:`Hist <histbook.hist.Hist>`
            histograms to combine (must have the same axes)
        """
        if any(not isinstance(x, Hist) for x in hists.values()):
            raise TypeError("only histograms can be grouped")

        axis = None
        hist = None
        for x in hists.values():
            hist = x
            if axis is None:
                axis = x._group + x._fixed + x._profile
            elif axis != x._group + x._fixed + x._profile:
                raise TypeError("histograms can only be grouped with the same axis specifications")
            
        if axis is None:
            raise ValueError("at least one histogram must be provided")

        if histbook.axis.groupby(by) in axis:
            raise ValueError("groupby({0}) already exists in these histograms; use hist.togroup(other) to add to a group".format(repr(by)))

        weight = None
        for x in hists.values():
            if weight is None:
                weight = x._weight
            elif x._weight is None:
                raise TypeError("histograms can only be grouped with the same weight specifications")
            elif weight != x._weight:
                weight = 1.0

        defs = {}
        for x in hists.values():
            defs.update(x._defs)

        out = Hist(*([histbook.axis.groupby(by)] + [x.relabel(x._original) for x in hist._group + hist._fixed + hist._profile]), weight=weight, defs=defs)
        out._content = {}
        for n, x in hists.items():
            out._content[n] = Hist._copycontent(x._content)
        return out

    def togroup(**hists):
        u"""
        Adds histograms to the :py:class:`groupby <histbook.axis.groupby>` that is the first axis.

        Histograms created with :py:meth:`Hist.group <histbook.hist.Hist.group>` have a first axis that is a :py:class:`groupby <histbook.axis.groupby>`.

        Keyword Arguments
        -----------------
        **hists : dict of str \u2192 :py:class:`Hist <histbook.hist.Hist>`
            histograms to add to the existing group
        """
        if len(self._group) == 0 or not isinstance(self._group[0], histbook.axis.groupby):
            raise ValueError("togroup can only be used on histograms whose first axis is a groupby")

        if any(not isinstance(x, Hist) for x in hists.values()):
            raise TypeError("only histograms can be grouped")

        for x in hists.values():
            if self._group[1:] + self._fixed + self._profile != x._group + x._fixed + x._profile:
                raise TypeError("histograms can only be grouped with the same axis specifications")

        def add(selfcontent, othercontent):
            assert isinstance(selfcontent, dict) and isinstance(othercontent, dict)
            for n in selfcontent:
                if n in othercontent:
                    if isinstance(selfcontent[n], numpy.ndarray):
                        selfcontent[n] += othercontent[n]
                    else:
                        add(selfcontent[n], othercontent[n])
            for n in othercontent:
                if n not in selfcontent:
                    selfcontent[n] = Hist._copycontent(othercontent[n])

        for n, x in hists.items():
            if n in self._content:
                add(self._content[n], other._content)
            else:
                self._content[n] = Hist._copycontent(other._content)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

from histbook.expr import *

class TestExpr(unittest.TestCase):
    def runTest(self):
        pass

    def test_const(self):
        self.assertEqual(Expr.parse("None"), Const(None))
        self.assertEqual(Expr.parse("True"), Const(True))
        self.assertEqual(Expr.parse("False"), Const(False))
        self.assertEqual(Expr.parse("1"), Const(1))
        self.assertEqual(Expr.parse("'hello'"), Const('hello'))
        self.assertEqual(Expr.parse("{}"), Const(set()))
        self.assertEqual(Expr.parse("{1, 2, 3}"), Const(set([1, 2, 3])))

    def test_name(self):
        self.assertEqual(Expr.parse("hello"), Name("hello"))

    def test_relation(self):
        self.assertEqual(Expr.parse("0 == x"), Relation("==", Const(0), Name("x")))
        self.assertEqual(Expr.parse("x == 0"), Relation("==", Const(0), Name("x")))
        self.assertEqual(Expr.parse("0 != x"), Relation("!=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("x != 0"), Relation("!=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("0 < x"), Relation("<", Const(0), Name("x")))
        self.assertEqual(Expr.parse("x < 0"), Relation("<", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("0 <= x"), Relation("<=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("x <= 0"), Relation("<=", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("0 > x"), Relation("<", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("x > 0"), Relation("<", Const(0), Name("x")))
        self.assertEqual(Expr.parse("0 >= x"), Relation("<=", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("x >= 0"), Relation("<=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("x in {}"), Relation("in", Name("x"), Const(set())))
        self.assertEqual(Expr.parse("x in {1, 2, 3}"), Relation("in", Name("x"), Const(set([1, 2, 3]))))
        self.assertEqual(Expr.parse("x not in {}"), Relation("not in", Name("x"), Const(set())))
        self.assertEqual(Expr.parse("x not in {1, 2, 3}"), Relation("not in", Name("x"), Const(set([1, 2, 3]))))

    def test_unary_plusminus(self):
        self.assertEqual(Expr.parse("+x"), Name("x"))
        self.assertEqual(Expr.parse("-x"), TimesDiv(-1, (Name("x"),), ()))
        self.assertEqual(Expr.parse("++x"), Name("x"))
        self.assertEqual(Expr.parse("--x"), Name("x"))
        self.assertEqual(Expr.parse("+++x"), Name("x"))
        self.assertEqual(Expr.parse("---x"), TimesDiv(-1, (Name("x"),), ()))
        self.assertEqual(Expr.parse("-2"), Const(-2))
        self.assertEqual(Expr.parse("-(2 + 2)"), Const(-4))

    def test_plusminus(self):
        xterm = TimesDiv(1, (Name("x"),), ())
        yterm = TimesDiv(1, (Name("y"),), ())
        zterm = TimesDiv(1, (Name("z"),), ())

        self.assertEqual(Expr.parse("x + y"), PlusMinus(0, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("y + x"), PlusMinus(0, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("x + 3"), PlusMinus(3, (xterm,), ()))
        self.assertEqual(Expr.parse("3 + x"), PlusMinus(3, (xterm,), ()))
        self.assertEqual(Expr.parse("(x + y) + z"), PlusMinus(0, (xterm, yterm, zterm), ()))
        self.assertEqual(Expr.parse("(z + y) + x"), PlusMinus(0, (xterm, yterm, zterm), ()))
        self.assertEqual(Expr.parse("x + (y + z)"), PlusMinus(0, (xterm, yterm, zterm), ()))
        self.assertEqual(Expr.parse("(x + y) + 3"), PlusMinus(3, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("(3 + y) + x"), PlusMinus(3, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("3 + (y + x)"), PlusMinus(3, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("x + (y + 3)"), PlusMinus(3, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("(x + y) + (z + 3)"), PlusMinus(3, (xterm, yterm, zterm), ()))
        self.assertEqual(Expr.parse("-(x + y) + z"), PlusMinus(0, (zterm,), (xterm, yterm)))
        self.assertEqual(Expr.parse("-(z + y) + x"), PlusMinus(0, (xterm,), (yterm, zterm)))
        self.assertEqual(Expr.parse("x + -(y + z)"), PlusMinus(0, (xterm,), (yterm, zterm)))
        self.assertEqual(Expr.parse("-(x + y) + 3"), PlusMinus(3, (), (xterm, yterm)))
        self.assertEqual(Expr.parse("-(3 + y) + x"), PlusMinus(-3, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("3 + -(y + x)"), PlusMinus(3, (), (xterm, yterm)))
        self.assertEqual(Expr.parse("x + -(y + 3)"), PlusMinus(-3, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("-(x + y) + (z + 3)"), PlusMinus(3, (zterm,), (xterm, yterm)))
        self.assertEqual(Expr.parse("(x + y) + -(z + 3)"), PlusMinus(-3, (xterm, yterm), (zterm,)))

        self.assertEqual(Expr.parse("x - y"), PlusMinus(0, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("y - x"), PlusMinus(0, (yterm,), (xterm,)))
        self.assertEqual(Expr.parse("x - 3"), PlusMinus(-3, (xterm,), ()))
        self.assertEqual(Expr.parse("3 - x"), PlusMinus(3, (), (xterm,)))
        self.assertEqual(Expr.parse("(x - y) - z"), PlusMinus(0, (xterm,), (yterm, zterm)))
        self.assertEqual(Expr.parse("(z - y) - x"), PlusMinus(0, (zterm,), (xterm, yterm)))
        self.assertEqual(Expr.parse("x - (y - z)"), PlusMinus(0, (xterm, zterm), (yterm,)))
        self.assertEqual(Expr.parse("(x - y) - 3"), PlusMinus(-3, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("(3 - y) - x"), PlusMinus(3, (), (xterm, yterm)))
        self.assertEqual(Expr.parse("3 - (y - x)"), PlusMinus(3, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("x - (y - 3)"), PlusMinus(3, (xterm,), (yterm,)))
        self.assertEqual(Expr.parse("(x - y) - (z - 3)"), PlusMinus(3, (xterm,), (yterm, zterm)))
        self.assertEqual(Expr.parse("-(x - y) - z"), PlusMinus(0, (yterm,), (xterm, zterm)))
        self.assertEqual(Expr.parse("-(z - y) - x"), PlusMinus(0, (yterm,), (xterm, zterm)))
        self.assertEqual(Expr.parse("x - -(y - z)"), PlusMinus(0, (xterm, yterm), (zterm,)))
        self.assertEqual(Expr.parse("-(x - y) - 3"), PlusMinus(-3, (yterm,), (xterm,)))
        self.assertEqual(Expr.parse("-(3 - y) - x"), PlusMinus(-3, (yterm,), (xterm,)))
        self.assertEqual(Expr.parse("3 - -(y - x)"), PlusMinus(3, (yterm,), (xterm,)))
        self.assertEqual(Expr.parse("x - -(y - 3)"), PlusMinus(-3, (xterm, yterm), ()))
        self.assertEqual(Expr.parse("-(x - y) - (z - 3)"), PlusMinus(3, (yterm,), (xterm, zterm)))
        self.assertEqual(Expr.parse("(x - y) - -(z - 3)"), PlusMinus(-3, (xterm, zterm), (yterm,)))

    def test_timesdiv(self):
        self.assertEqual(Expr.parse("x * y"), TimesDiv(1, (Name("x"), Name("y")), ()))
        self.assertEqual(Expr.parse("y * x"), TimesDiv(1, (Name("x"), Name("y")), ()))
        self.assertEqual(Expr.parse("3 * x"), TimesDiv(3, (Name("x"),), ()))
        self.assertEqual(Expr.parse("x * 3"), TimesDiv(3, (Name("x"),), ()))
        self.assertEqual(Expr.parse("(x * y) * z"), TimesDiv(1, (Name("x"), Name("y"), Name("z")), ()))
        self.assertEqual(Expr.parse("x * (y * z)"), TimesDiv(1, (Name("x"), Name("y"), Name("z")), ()))
        self.assertEqual(Expr.parse("(x * 3) * y"), TimesDiv(3, (Name("x"), Name("y")), ()))
        self.assertEqual(Expr.parse("x * (3 * y)"), TimesDiv(3, (Name("x"), Name("y")), ()))

        self.assertEqual(Expr.parse("x / y"), TimesDiv(1, (Name("x"),), (Name("y"),)))
        self.assertEqual(Expr.parse("y / x"), TimesDiv(1, (Name("y"),), (Name("x"),)))
        self.assertEqual(Expr.parse("3 / x"), TimesDiv(3, (), (Name("x"),)))
        self.assertEqual(Expr.parse("x / 3"), TimesDiv(1.0/3.0, (Name("x"),), ()))
        self.assertEqual(Expr.parse("(x / y) / z"), TimesDiv(1, (Name("x"),), (Name("y"), Name("z"))))
        self.assertEqual(Expr.parse("x / (y / z)"), TimesDiv(1, (Name("x"), Name("z")), (Name("y"),)))
        self.assertEqual(Expr.parse("(x / 3) / y"), TimesDiv(1.0/3.0, (Name("x"),), (Name("y"),)))
        self.assertEqual(Expr.parse("x / (3 / y)"), TimesDiv(1.0/3.0, (Name("x"), Name("y")), ()))

    def test_distributive(self):
        self.assertEqual(Expr.parse("a * (x + y)"), PlusMinus(0, (TimesDiv(1, (Name("a"), Name("x")), ()), TimesDiv(1, (Name("a"), Name("y")), ())), ()))
        self.assertEqual(Expr.parse("(x + y) * a"), PlusMinus(0, (TimesDiv(1, (Name("a"), Name("x")), ()), TimesDiv(1, (Name("a"), Name("y")), ())), ()))
        self.assertEqual(Expr.parse("a * (x + 3)"), PlusMinus(0, (TimesDiv(1, (Name("a"), Name("x")), ()), TimesDiv(3, (Name("a"),), ())), ()))
        self.assertEqual(Expr.parse("(x + 3) * a"), PlusMinus(0, (TimesDiv(1, (Name("a"), Name("x")), ()), TimesDiv(3, (Name("a"),), ())), ()))

        self.assertEqual(Expr.parse("a * (x + y) + a*x + a*y"), PlusMinus(0, (TimesDiv(2, (Name("a"), Name("x")), ()), TimesDiv(2, (Name("a"), Name("y")), ())), ()))
        self.assertEqual(Expr.parse("a * (x + y) - 2*a*x"), PlusMinus(0, (TimesDiv(1, (Name("a"), Name("y")), ()),), (TimesDiv(1, (Name("a"), Name("x")), ()),)))

        # simplifies (because it stays within the ring)
        self.assertEqual(Expr.parse("(x + y) / a"), PlusMinus(0, (TimesDiv(1, (Name("x"),), (Name("a"),)), TimesDiv(1, (Name("y"),), (Name("a"),))), ()))
        self.assertEqual(Expr.parse("(x + 3) / a"), PlusMinus(0, (TimesDiv(1, (Name("x"),), (Name("a"),)), TimesDiv(3, (), (Name("a"),))), ()))

        # does not simplify (because division of a sum puts it into the field)
        self.assertEqual(Expr.parse("a / (x + y)"), TimesDiv(1, (Name("a"),), (PlusMinus(0, (TimesDiv(1, (Name("x"),), ()), TimesDiv(1, (Name("y"),), ())), ()),)))
        self.assertEqual(Expr.parse("a / (x + 3)"), TimesDiv(1, (Name("a"),), (PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()),)))

        self.assertEqual(Expr.parse("(2 - 2 - x) / a"), TimesDiv(-1, (Name("x"),), (Name("a"),)))
        self.assertEqual(Expr.parse("a / (2 - 2 - x)"), TimesDiv(-1, (Name("a"),), (Name("x"),)))

    def test_cancellation(self):
        self.assertEqual(Expr.parse("x - x"), Const(0))
        self.assertEqual(Expr.parse("x + x - 2*x"), Const(0))
        self.assertEqual(Expr.parse("a * (x + y) - a*x - a*y"), Const(0))
        self.assertEqual(Expr.parse("a * (x + y) - a*x"), TimesDiv(1, (Name("a"), Name("y")), ()))
        self.assertEqual(Expr.parse("a * (x + y)/y - a*x/y"), Name("a"))

        self.assertEqual(Expr.parse("(x + x*x)/x - x"), Const(1))
        self.assertEqual(Expr.parse("(x + x*x)/x - 1"), Name("x"))
        self.assertEqual(Expr.parse("x - (x + x*x)/x"), Const(-1))
        self.assertEqual(Expr.parse("1 - (x + x*x)/x"), TimesDiv(-1, (Name("x"),), ()))

    def test_lowintpower(self):
        self.assertEqual(Expr.parse("x**1"), Name("x"))
        self.assertEqual(Expr.parse("x**2"), TimesDiv(1, (Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("x**3"), TimesDiv(1, (Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("x**4"), TimesDiv(1, (Name("x"), Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-x)**1"), TimesDiv(-1, (Name("x"),), ()))
        self.assertEqual(Expr.parse("(-x)**2"), TimesDiv(1, (Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-x)**3"), TimesDiv(-1, (Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-x)**4"), TimesDiv(1, (Name("x"), Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(2*x)**1"), TimesDiv(2, (Name("x"),), ()))
        self.assertEqual(Expr.parse("(2*x)**2"), TimesDiv(4, (Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(2*x)**3"), TimesDiv(8, (Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(2*x)**4"), TimesDiv(16, (Name("x"), Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-2*x)**1"), TimesDiv(-2, (Name("x"),), ()))
        self.assertEqual(Expr.parse("(-2*x)**2"), TimesDiv(4, (Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-2*x)**3"), TimesDiv(-8, (Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(-2*x)**4"), TimesDiv(16, (Name("x"), Name("x"), Name("x"), Name("x")), ()))
        self.assertEqual(Expr.parse("(x + 3)**1"), PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()))
        self.assertEqual(Expr.parse("(x + 3)**2"), TimesDiv(1, (PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()), PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ())), ()))

        self.assertEqual(Expr.parse("x**-1"), TimesDiv(1, (), (Name("x"),)))
        self.assertEqual(Expr.parse("x**-2"), TimesDiv(1, (), (Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("x**-3"), TimesDiv(1, (), (Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("x**-4"), TimesDiv(1, (), (Name("x"), Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-x)**-1"), TimesDiv(-1, (), (Name("x"),)))
        self.assertEqual(Expr.parse("(-x)**-2"), TimesDiv(1, (), (Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-x)**-3"), TimesDiv(-1, (), (Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-x)**-4"), TimesDiv(1, (), (Name("x"), Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(2*x)**-1"), TimesDiv(1.0/2.0, (), (Name("x"),)))
        self.assertEqual(Expr.parse("(2*x)**-2"), TimesDiv(1.0/4.0, (), (Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(2*x)**-3"), TimesDiv(1.0/8.0, (), (Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(2*x)**-4"), TimesDiv(1.0/16.0, (), (Name("x"), Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-2*x)**-1"), TimesDiv(-1.0/2.0, (), (Name("x"),)))
        self.assertEqual(Expr.parse("(-2*x)**-2"), TimesDiv(1.0/4.0, (), (Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-2*x)**-3"), TimesDiv(-1.0/8.0, (), (Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(-2*x)**-4"), TimesDiv(1.0/16.0, (), (Name("x"), Name("x"), Name("x"), Name("x"))))
        self.assertEqual(Expr.parse("(x + 3)**-1"), TimesDiv(1, (), (PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()),)))
        self.assertEqual(Expr.parse("(x + 3)**-2"), TimesDiv(1, (), (PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()), PlusMinus(3, (TimesDiv(1, (Name("x"),), ()),), ()))))

    def test_binop(self):
        pass

    def test_logicalnot(self):
        self.assertEqual(Expr.parse("not x == 0"), Relation("!=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("not x != 0"), Relation("==", Const(0), Name("x")))
        self.assertEqual(Expr.parse("not x < 0"), Relation("<=", Const(0), Name("x")))
        self.assertEqual(Expr.parse("not x <= 0"), Relation("<", Const(0), Name("x")))
        self.assertEqual(Expr.parse("not x > 0"), Relation("<=", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("not x >= 0"), Relation("<", Name("x"), Const(0)))
        self.assertEqual(Expr.parse("not x in {}"), Relation("not in", Name("x"), Const(set())))
        self.assertEqual(Expr.parse("not x not in {}"), Relation("in", Name("x"), Const(set())))

        self.assertEqual(Expr.parse("not p"), Predicate("p", positive=False))

    def test_logicals(self):
        self.assertEqual(Expr.parse("a and b"), LogicalAnd(Predicate("a"), Predicate("b")))
        self.assertEqual(Expr.parse("a or b"), LogicalOr(LogicalAnd(Predicate("a")), LogicalAnd(Predicate("b"))))
        self.assertEqual(Expr.parse("(a and b) or (c and d)"), LogicalOr(LogicalAnd(Predicate("a"), Predicate("b")), LogicalAnd(Predicate("c"), Predicate("d"))))
        self.assertEqual(Expr.parse("(a or b) and (c or d)"), LogicalOr(LogicalAnd(Predicate("a"), Predicate("c")), LogicalAnd(Predicate("a"), Predicate("d")), LogicalAnd(Predicate("b"), Predicate("c")), LogicalAnd(Predicate("b"), Predicate("d"))))
        self.assertEqual(Expr.parse("(a or b) and c"), LogicalOr(LogicalAnd(Predicate("a"), Predicate("c")), LogicalAnd(Predicate("b"), Predicate("c"))))
        self.assertEqual(Expr.parse("c and (a or b)"), LogicalOr(LogicalAnd(Predicate("a"), Predicate("c")), LogicalAnd(Predicate("b"), Predicate("c"))))
        self.assertEqual(Expr.parse("c or (a or b)"), LogicalOr(LogicalAnd(Predicate("a")), LogicalAnd(Predicate("b")), LogicalAnd(Predicate("c"))))

    def test_logical_negations(self):
        self.assertEqual(Expr.parse("not (a and b)"), LogicalOr(LogicalAnd(Predicate("a", False)), LogicalAnd(Predicate("b", False))))
        self.assertEqual(Expr.parse("not (a or b)"), LogicalAnd(Predicate("a", False), Predicate("b", False)))
        self.assertEqual(Expr.parse("not (x == 123 and x == 999)"), LogicalOr(LogicalAnd(Relation("!=", Const(123), Name("x"))), LogicalAnd(Relation("!=", Const(999), Name("x")))))
        self.assertEqual(Expr.parse("not (x == 123 or x == 999)"), LogicalAnd(Relation("!=", Const(123), Name("x")), Relation("!=", Const(999), Name("x"))))

    def test_function(self):
        self.assertEqual(Expr.parse("sqrt(x)"), Call("sqrt", Name("x")))
